import codecs
from pathlib import Path

import PySimpleGUI as sg

import criticus.py.edit_settings as es
import criticus.py.custom_popups as cp
from criticus.py.export_to_docx.xml_to_docx import export_xml_to_docx
from criticus.py import edit_settings as es


def unescape_string(text: str):
    return codecs.decode(text, 'unicode-escape')

def validate_form(values):
    for value in values:
        if values[value] == '':
            print(f'{values[value]=}')
            cp.ok('Please fill in all input fields', 'No Blanks, Please')
            return
        return True

def update_settings(settings: dict, values: dict):
    settings['text_wits_separator'] = unescape_string(values['text_wits_separator'])
    settings['rdg_n_text_separator'] = unescape_string(values['rdg_n_text_separator'])
    settings['words_per_line'] = values[ 'words_per_line']
    settings['text_bold'] = values['text_bold']
    settings['reformatted_xml_dir'] = Path(values['xml_filename']).parent.as_posix()
    es.save_settings(settings)

def export(settings, values: dict, collapse_regularized: bool = False, use_custom_template: bool = None, add_suffix: bool = False):
    if not validate_form(values):
        return
    update_settings(settings, values)
    # try:
    saved_file = export_xml_to_docx(values['xml_filename'], collapse_regularized, use_custom_template, add_suffix)
    if saved_file:
        cp.ok(f'Collation exported to {saved_file}', 'Success!')
    # except Exception as e:
    #     cp.ok(f'{e}', 'Failed to Export')

def layout(settings: dict):
    custom_template_path = es.get_settings()['custom_template_path']
    if  custom_template_path != '':
        template_parent = Path(custom_template_path).parent.as_posix()
    else:
        template_parent = None

    options_frame = [
        [
            sg.Checkbox('Use Custom Template', key='use_custom_template', enable_events=True), sg.Input(custom_template_path, key='custom_template_path', disabled=True), sg.FileBrowse(initial_folder=template_parent)
        ],
        [
            sg.T('Reading Text and Witnesses Separator: '), 
            sg.I(settings.get('text_wits_separator', ' // '), key='text_wits_separator')
        ],
        [
            sg.T('Reading ID and Reading Text Separator: '),
            sg.I(settings.get('rdg_n_text_separator', '\t'), key='rdg_n_text_separator')
        ],
        [
            sg.T('Basetext Words Per Line: '), 
            sg.Spin([i for i in range(1, 20)], initial_value=settings.get('words_per_line', 10), key='words_per_line')
        ],
        [
            sg.Check('Make Reading Text Bold', default=settings.get('text_bold', False), key='text_bold')
        ],
        [
            sg.Checkbox('Collapse regularized readings to parent reading', key='collapse_regularized'),
            sg.Checkbox('Add "r" suffix to each witness moved to its parent reading', key='add_r_suffix')
        ]
    ]
    return [
        [
            sg.T('Collation File: '), 
            sg.I('', key='xml_filename'), 
            sg.FileBrowse(file_types=(('XML Files', '*.xml'),), initial_folder=settings.get('reformatted_xml_dir'))
        ],
        [
            sg.Frame('Export Options', options_frame)
        ],
        [
            sg.Button('Export'), sg.Button('Close')
        ]
    ]

def export_to_docx(font, icon):
    settings = es.get_settings()
    window = sg.Window('Export Collation to DOCX', layout(settings), icon=icon, font=font)
    while True:
        event, values = window.read()
        if event in [None, sg.WIN_CLOSED, 'Close']:
            break
        elif event == 'Export' and not values['use_custom_template']:
            export(settings, values, values['collapse_regularized'], None, values['add_r_suffix'])

        elif event == 'Export' and values['use_custom_template'] is True and values['custom_template_path'] != '':
            export(settings, values, values['collapse_regularized'], values['custom_template_path'], values['add_r_suffix'])
            es.edit_settings('custom_template_path', values['custom_template_path'])

        elif event == 'use_custom_template':
            window['custom_template_path'].update(disabled=not values['use_custom_template'])
    window.close()



import json
import lxml.etree as et

import PySimpleGUI as sg

from criticus.py.tei2json.to_json import verse_to_dict, save_tx
from criticus.py.tei2json.from_tei import (get_file, pre_parse_cleanup,
                      add_underdot_to_unclear_letters,
                      parse, remove_unclear_tags,
                      tei_ns, get_verse_as_tuple)
import criticus.py.custom_popups as cp
# pylint: disable=no-member
#########
def get_siglum_from_user() -> str:
    msg = '''criticus could not find the siglum.
Please enter a witness ID: '''
    layout = [[sg.T(msg)],
              [sg.I('', key='siglum')],
              [sg.B('Submit')]]
    window = sg.Window('Provide a Siglum', layout)
    siglum = ''
    while True:
        event, values = window.read()
        if event in [None, sg.WINDOW_CLOSED]:
            siglum = values['siglum']
            break
        elif event == 'Submit' and values['siglum'] != '':
            siglum = values['siglum']
            break
    window.close()
    return siglum
#########

def get_siglum(root: et._Element) -> str:
    titles = root.xpath('//tei:title', namespaces={'tei': tei_ns})
    for title in titles:
        if title.get('n'):
            siglum = title.get('n')
            siglum = siglum.replace('-ns', '')
            break
    else:
        siglum = ''
        while siglum == '':
            siglum = get_siglum_from_user()
    return siglum

def get_hands(root: et._Element) -> list:
    rdgs = root.xpath('//tei:rdg', namespaces={'tei': tei_ns})
    hands = []
    for rdg in rdgs:
        if rdg.get('hand') and rdg.get('hand') not in hands:
            hands.append(rdg.get('hand'))
    if hands == []:
        hands = ['firsthand']
    return hands

def tei_to_json(tei: str, output_dir, single_verse: str):
    text = get_file(tei)
    text = pre_parse_cleanup(text)
    parsed, root = parse(text)
    if not parsed:
        cp.ok(f'Failed to parse XML. See error:\n{root}', title='Bummer...')
        return False
    add_underdot_to_unclear_letters(root)
    text = et.tostring(root, encoding='unicode')
    text = remove_unclear_tags(text)
    _, root = parse(text)
    hands = get_hands(root)
    siglum = get_siglum(root)
    output_dir = f'{output_dir}/{siglum}'
    metadata = {'id': siglum, 'siglum': siglum}
    verses = root.xpath(f'//tei:ab', namespaces={'tei': tei_ns})
    for verse in verses:
        ref = verse.get('n')
        if single_verse != '' and single_verse != ref:
            continue
        witnesses = get_verse_as_tuple(verse, hands=hands)
        verse_as_dict = verse_to_dict(siglum, ref, witnesses)
        save_tx(verse_as_dict, ref, output_dir)
    if output_dir:
        f = f'{output_dir}/metadata.json'
    else:
        f = 'metadata.json'
    with open(f, 'w', encoding='utf-8') as file:
        json.dump(metadata, file, ensure_ascii=False)
    return True



from copy import deepcopy
import os
from lxml import etree as et

from natsort import natsorted
import PySimpleGUI as sg

import criticus.py.edit_settings as es
import criticus.py.custom_popups as cp
import criticus.py.reformat_collation.itsee_to_open_cbgm as itoc

#pylint: disable=no-member

def get_verse_file(f, output_dir):
    parser = et.XMLParser(remove_blank_text=True, recover=True)
    tree = et.parse(f'{output_dir}/{f}', parser=parser)
    root = tree.getroot()
    ns = root.nsmap
    all_ab_elems = root.findall('ab', ns)
    return deepcopy(all_ab_elems)

def combine_verses(starting_string: str, output_dir, main_dir, already_formatted: bool, title_stmt: str, publication_stmt: str):
    tree = et.parse(f'{main_dir}/resources/template.xml')
    root = tree.getroot()
    files = os.listdir(output_dir)
    files = natsorted(files)
    failed = []
    for f in files:
        if f.startswith(starting_string):
            try:
                all_ab_elems = get_verse_file(f, output_dir)
                for ab in all_ab_elems:
                    root.append(ab)
            except Exception as e:
                failed.append(f'{f}: {e}')
    if len(failed) > 0:
        cp.listbox(
            'The following files failed to parse and were skipped.',
            failed, 'Some files failed to combine'
            )
    if already_formatted:
        itoc.add_tei_header(tree, title_stmt, publication_stmt)
    return tree

def combine_xml_files_interface(main_dir, font, icon):
    titlestmt_tip = '''E.g. "Collation of Romans with Emphasis on Lectionary Witnesses"'''
    pub_tip = '''E.g. "This is an open access work licensed under a Creative Commons Attribution 4.0 International license."'''
    settings = es.get_settings()
    layout = [
        [sg.Text('Select the folder that contains the individual XML files to be combined')],
        [sg.Text('Folder:'), sg.Input(settings['ce_output_dir'], key='output_dir'), sg.FolderBrowse(initial_folder=settings['ce_output_dir'])],
        [sg.Text('Combine all files that start with:'), sg.Stretch(), sg.Input('', key='starts_with')],
        [sg.Checkbox('Files have already been reformatted', key='already_reformatted', enable_events=True)],
        [sg.T('Collation Title:'), sg.Input('untitled', key='title_stmt', expand_x=True, tooltip=titlestmt_tip, disabled=True)],
        [sg.T('Publication Statement:'), sg.Input('unspecified', key='publication_stmt', expand_x=True, tooltip=pub_tip, disabled=True)],
        [sg.Button('Combine XML Files'), sg.Button('Cancel')]
    ]
    window = sg.Window('Combine XML Files', layout, icon=icon, font=font)
    while True:
        event, values = window.read()
        if event in [sg.WINDOW_CLOSED, 'Cancel']:
            break
        elif event == 'Combine XML Files':
            if values['output_dir'] == '' or values['starts_with'] == '':
                sg.popup_quick_message('First select a folder and at least one "starts with" character')
                continue
            try:
                tree = combine_verses(values['starts_with'], values['output_dir'], main_dir, values['already_reformatted'], values['title_stmt'], values['publication_stmt'])
                es.edit_settings('ce_output_dir', values['output_dir'])
            except Exception as e:
                cp.ok(f'Failed to combine files because:\n" {e} "\n\
Double-check that the selected folder contains XML files and that\n\
these match the given "starts with" characters', 'Bummer')
                continue
            saved_name = sg.popup_get_file('', no_window=True, file_types=(("XML Files", "*.xml"),), save_as=True, initial_folder=values['output_dir'])
            if saved_name:
                tree.write(saved_name, encoding='utf-8', xml_declaration=True)
                settings['ce_output_dir'] = values['output_dir']
                cp.ok(f'Files have been combined and saved to {saved_name}', 'Combined!')
                break
        elif event == 'already_reformatted':
            window['title_stmt'].update(disabled=not values['already_reformatted'])
            window['publication_stmt'].update(disabled=not values['already_reformatted'])
    window.close()
    return False



from pathlib import Path
import platform
import PySimpleGUI as sg
import criticus.py.edit_settings as es
from criticus.py.txt2json.convert_text_to_json import convert_text_to_json as t2j

#pylint: disable=no-member
def disable_from_and_to(window: sg.Window, switch: bool, values):
    window['range_from'].update(disabled=switch)
    window['range_to'].update(disabled=switch)
    disable_buttons(window, values)
    
def disable_siglum_and_prefix(window: sg.Window, switch: bool, values):
    window['siglum_input'].update(disabled=switch)
    window['ref_prefix_input'].update(disabled=switch)
    disable_buttons(window, values)

def browse_for_output_dir(window: sg.Window, output_dir):
    window['output_dir_input'].update(output_dir)
    es.edit_settings('ce_repo_dir', output_dir)

def disable_buttons(window: sg.Window, values: dict):
    if ((values['all_verses_in_file'] is False and values['range_of_verses'] is False) 
    or (values['manual'] is False and values['auto'] is False) 
    or values['output_dir_input'] in ['', None]):
       window['convert_dir'].update(disabled=True) 
       window['convert_file'].update(disabled=True)
    else:
        window['convert_dir'].update(disabled=False) 
        window['convert_file'].update(disabled=False)

def convert_file(values: dict, icon):
    settings = es.get_settings()
    filename = sg.popup_get_file('', no_window=True, initial_folder=settings['tx_dir'], file_types=(('Plain Text Files', '*.txt'),))
    if not filename:
        return
    f = Path(filename)
    f = f.parent.absolute().as_posix()
    es.edit_settings('tx_dir', f)
    t2j(
        filename, output_dir=values['output_dir_input'], 
        convert_all=values['all_verses_in_file'],
        reference_prefix=values['ref_prefix_input'],
        auto=values['auto'], verse_from=values['range_from'],
        verse_to=values['range_to'], siglum=values['siglum_input']
    )
    sg.popup_ok('Done!', title='Text File Converted', icon=icon)

def convert_dir(values: dict, icon):
    settings = es.get_settings()
    folder = sg.popup_get_folder('', no_window=True, initial_folder=settings['tx_dir'], icon=icon)
    if folder:
        es.edit_settings('tx_dir', folder)
        folder = Path(folder)
        for f in folder.iterdir():
            if f.name.endswith('.txt'):
                t2j(
                    f, 
                    output_dir=values['output_dir_input'], 
                    convert_all=True,
                    reference_prefix=values['ref_prefix_input'],
                    auto=values['auto'], 
                    verse_from=values['range_from'],
                    verse_to=values['range_to'], 
                    siglum=values['siglum_input']
                )
        sg.popup_ok('Done!', title='Text File Converted', icon=icon)


def txt_to_json(font: tuple, icon):
    settings = es.get_settings()
    if platform.system() == 'Windows':
        space = sg.T('')
        output_folder_elem = sg.Input(default_text=settings['ce_repo_dir'], disabled=True, key='output_dir_input')
    else:
        space = sg.T('               ')
        output_folder_elem = sg.Input(default_text=settings['ce_repo_dir'], disabled=True, key='output_dir_input', size=(30, 1))
        
    frame_prepare_all_or_rage = [
        [sg.Radio('All verses in file ', group_id='all_or_range', key='all_verses_in_file', enable_events=True)],
        [sg.Radio('Range of verses ', group_id='all_or_range', key='range_of_verses', enable_events=True),
                sg.T('From'), sg.Input(key='range_from', disabled=True), sg.T('To'), sg.Input(key='range_to', disabled=True)],
    ]
    frame_ref_format = [
        [sg.Radio('Manual ', group_id='ref_prefix', enable_events=True, key='manual'), 
                sg.T('Siglum'), sg.Input(key='siglum_input', disabled=True), 
                sg.T('Unit prefix'), sg.Input(key='ref_prefix_input', disabled=True)],
        [sg.Radio('Auto from file name ', group_id='ref_prefix', enable_events=True, key='auto')],
    ]

    win_txt_to_json = [
        [sg.Button('Back', key='exit'), sg.Stretch()],
        [sg.Frame('Units to Convert', frame_prepare_all_or_rage)],
        [sg.Frame('Transcription Info', frame_ref_format)],
        [sg.T('Output Directory '), output_folder_elem, sg.Button('Browse')],
        [sg.Button('Convert File', key='convert_file', disabled=True), space,
                sg.Button('Convert Directory', key='convert_dir', disabled=True)],
    ]

    window = sg.Window('Convert Plain Text to JSON', win_txt_to_json, font=font, icon=icon)

###########################################################
###########################################################

    while True:
        event, values = window.read()

        if event == sg.WIN_CLOSED:
            break

        elif event == 'range_of_verses':
            disable_from_and_to(window, False, values)

        elif event == 'all_verses_in_file':
            disable_from_and_to(window, True, values)

        elif event == 'auto':
            disable_siglum_and_prefix(window, True, values)

        elif event == 'manual':
            disable_siglum_and_prefix(window, False, values)

        elif event == 'exit':
            break

        elif event == 'Browse':
            output_dir = sg.popup_get_folder('', no_window=True, default_path=settings['ce_repo_dir'])
            browse_for_output_dir(window, output_dir)

        elif event == 'convert_file':
            convert_file(values, icon)

        elif event == 'convert_dir':
            convert_dir(values, icon)

    window.close()
    return False



import platform

import PySimpleGUI as sg

import criticus.py.edit_settings as es
import criticus.py.custom_popups as cp
from criticus.py.tei2json.tei_to_json import tei_to_json as t2j
from criticus.py.tei2json.edit_regex import edit_regex

# pylint: disable=no-member
def get_siglum_from_user(msg: str, title: str, icon) -> str:
    layout = [[sg.T(msg)], [sg.I('', key='input')], [sg.B('Submit')]]
    window = sg.Window(title, layout, icon=icon)
    siglum = ''
    while True:
        event, values = window.read()
        if event in [None, sg.WIN_CLOSED]:
            break
        elif event == 'Submit' and values['input'] != '':
            siglum = values['input']
            break
    window.close()
    return siglum

def get_space():
    return sg.T('          ')

def no_space():
    return sg.T('')

def layout(settings: dict):
    if platform.system() == 'Windows':
        space = no_space
        input_elem = sg.I('', key='tei_input')
        output_folder_elem = sg.I(settings['ce_repo_dir'], key='output_dir')
    else:
        space = get_space
        input_elem = sg.I('', key='tei_input', size=(30, 1))
        output_folder_elem = sg.I(settings['ce_repo_dir'], key='output_dir', size=(30, 1))
    menu = [['Settings', ['Add/Remove Regular Expressions']]]
    input_frame = [
        [input_elem, sg.FileBrowse(initial_folder=settings['tei_dir'], file_types=(('XML Files', '*.xml'), ))],
        [sg.Radio('Convert All Verses', 'all_or_one', key='all', enable_events=True)],
        [sg.Radio('Convert One Verse ', 'all_or_one', key='one', enable_events=True), sg.T('Reference'), sg.I('', key='single_ref', disabled=True, enable_events=True)]
    ]
    output_frame = [
        [output_folder_elem, sg.FolderBrowse(initial_folder=settings['ce_repo_dir'])]
    ]
    return [
        [sg.Menu(menu)],
        [sg.Frame('TEI Transcription File', input_frame)],
        [sg.Frame('Output Folder', output_frame)],
        [sg.B('Convert', disabled=True, key='convert'), space(), sg.B('Cancel', key='exit')]
        ]

def save_settings(values):
    es.edit_settings('tei_dir', values['tei_input'])
    es.edit_settings('ce_repo_dir', values['output_dir'])

def enable_ui_elems(values: dict, window: sg.Window):
    if values['one'] is True:
        window['single_ref'].update(disabled=False)
    else:
        window['single_ref'].update(disabled=True)
        window['single_ref'].update('')
    if ((values['all'] is False and values['one'] is False)
    or (values['one'] is True and values['single_ref'] == '')):
        window['convert'].update(disabled=True)
    else:
        window['convert'].update(disabled=False)

def tei_to_json(font: tuple, icon):
    settings = es.get_settings()
    window = sg.Window('Convert TEI to JSON', layout(settings), font=font, icon=icon)

    while True:
        event, values = window.read()

        if event in [sg.WIN_CLOSED, 'exit', None]:
            break

        elif event in ['all', 'one', 'single_ref']:
            enable_ui_elems(values, window)

        elif event == 'convert':
            if values['tei_input'] == '' or values['output_dir'] == '':
                continue
            else:
                save_settings(values)
                try:
                    result = t2j(values['tei_input'], values['output_dir'], single_verse=values['single_ref'])
                    if result:
                        cp.ok(f'JSON transcription files saved to {values["output_dir"]}', 'Success!')
                        print('success!')
                    else:
                        continue
                except Exception as e:
                    # print('conversion failed')
                    cp.ok(f'Conversion failed. Talk to David. See error below:\n{e}', 'Bummer...')
        elif event == 'Add/Remove Regular Expressions':
            edit_regex(icon)

    window.close()
    return False



import http.server as server
import threading
import shutil
import socketserver
from pathlib import Path
import os
import platform
import sys


def start_tei_server(tei_repo_dir, main_dir, httpd, PORT):
    tei_repo_dir = Path(tei_repo_dir).as_posix()
    main_dir = Path(main_dir).as_posix()
    try:
        shutil.copy(f'{main_dir}/tei_transcription.xsl', tei_repo_dir)
    except:
        pass
    print('changing to tei dir')
    os.chdir(tei_repo_dir)
    print(f'Server started at localhost: {PORT}')
    sys.stdout = sys.stderr = open(os.devnull, "w")
    httpd.serve_forever(1)


def launch_tei_viewer(tei_repo_dir, main_dir) -> socketserver.TCPServer:
    PORT = 8011
    handler = server.SimpleHTTPRequestHandler
    httpd = socketserver.TCPServer(('', PORT), handler)
    t = threading.Thread(target=start_tei_server, args=[tei_repo_dir, main_dir, httpd, PORT], daemon=True)
    t.start()
    try:
        if platform.system() == 'Windows':
            from subprocess import Popen
            Popen(f'start firefox http://localhost:{PORT}', shell=True)
        else:
            import webbrowser
            webbrowser.get('firefox').open(f'http://localhost:{PORT}/')
    except Exception as e:
        print(f'Could not open browser because:\n{e}')
    return httpd



import json
import os
from pathlib import Path
import platform
import subprocess
from typing import Dict, List
import webbrowser

from natsort import natsorted
import PySimpleGUI as sg

import criticus.py.edit_settings as es

# pylint: disable=no-member
def get_config(fn = None) -> dict:
    if not fn:
        settings = es.get_settings()
        fn = settings['ce_config_fn']
    try:
        with open(fn, 'r', encoding='utf-8') as c:
            config = json.load(c)
            if not config.get('excluded_witnesses'):
                config['excluded_witnesses'] = []
            return config
    except:
        sg.popup_quick_message('No config file found.\nBrowse for the Config File')
        return {}

def save_config(config: dict, fn: str):
    config['witnesses'] = sort_by_ga(config['witnesses'])
    config['excluded_witnesses'] = sort_by_ga(config['excluded_witnesses'])
    with open(fn, 'w', encoding='utf-8') as f:
        json.dump(config, f, ensure_ascii=False, indent=4)

def layout():
    settings = es.get_settings()
    i_size = (30, 1)
    config = get_config()
    if platform.system() == 'Windows':
        launch_ce = sg.B('Start Collation Editor')
    else:
        # launch_ce = sg.T('')
        launch_ce = sg.B('Start Collation Editor')

    settings_frame = [
        [sg.T('Project Title'), sg.I(config.get('name', ''), size=i_size, key='name')],
        [sg.T('Basetext'), sg.I(config.get('base_text', ''), size=i_size, key='basetext')],
        [sg.T('Included'), sg.Listbox(config.get('witnesses', []), select_mode=sg.LISTBOX_SELECT_MODE_EXTENDED, key='witnesses', expand_x=True, size=(None, 12))],
        [sg.B('Add Witness', bind_return_key=True), sg.I('', key='wit_to_add'), sg.T(''), sg.B('Move to Excluded')],
        [sg.T('Excluded'), sg.Listbox(config.get('excluded_witnesses', []), select_mode=sg.LISTBOX_SELECT_MODE_EXTENDED, key='excluded', expand_x=True, size=(None, 12))],
        [sg.B('Move to Included'), sg.B('Delete Selected')]
    ]

    return [
        [sg.T('Collation Config File'), sg.I(settings['ce_config_fn'], key='config_fn'), sg.FileBrowse(file_types=(('JSON Files', '*.json'),)), sg.B('Update')],
        [sg.Frame('Collation Configuration', settings_frame, border_width=4, expand_x=True, expand_y=True)],
        [launch_ce, sg.T(''), sg.B('Done', key='exit')],
    ]

def edit_config(values):
    config = get_config(values['config_fn'])
    if values['name'] != '':
        config['name'] = values['name']
    if values['basetext'] != '':
        config['base_text'] = values['basetext']
    save_config(config, values['config_fn'])
    sg.popup_quick_message('Configuration File Updated')

def sort_by_ga(wits: List[str]):
    papyri = []
    majuscules = []
    minuscules = []
    lectionaries = []
    editions = []
    for wit in wits:
        if wit.lower().startswith('p'):
            papyri.append(wit)
        elif wit.startswith('0'):
            majuscules.append(wit)
        elif wit[0].isdigit():
            minuscules.append(wit)
        elif wit.lower().startswith('l'):
            lectionaries.append(wit)
        else:
            editions.append(wit)
    return natsorted(papyri) + natsorted(majuscules) + natsorted(minuscules) + natsorted(lectionaries) + natsorted(editions)
  
def remove_from_config(config: dict, key, item_to_remove: str):
    new_list = []
    for wit in config[key]:
        if wit != item_to_remove and wit not in new_list:
            new_list.append(wit)
    config[key] = new_list
    return config

def update_window(window: sg.Window, values):
    config = get_config(values['config_fn'])
    if not config:
        return
    window['name'].update(config['name'])
    window['basetext'].update(config['base_text'])
    window['witnesses'].update(config['witnesses'])
    window['excluded'].update(config['excluded_witnesses'])
    window['wit_to_add'].update('')

def add_witness(values, window):
    if values['wit_to_add'] == '':
        return
    if values['config_fn'] == '':
        return
    config = get_config(values['config_fn'])
    try:
        config['witnesses'].append(values['wit_to_add'])
    except:
        sg.popup_quick_message('Are you sure that is the right config.json file?\nIt did not work.')
        return
    config = remove_from_config(config, 'excluded_witnesses', values['wit_to_add'])
    save_config(config, values['config_fn'])
    update_window(window, values)

def remove_witnesses(values, window):
    if values['witnesses'] == []:
        return
    config = get_config(values['config_fn'])
    for wit in values['witnesses']:
        if wit not in config['excluded_witnesses']:
            config['excluded_witnesses'].append(wit)
        config = remove_from_config(config, 'witnesses', wit)
    save_config(config, values['config_fn'])
    update_window(window, values)

def include_selected(values: dict, window: sg.Window):
    if values['excluded'] == []:
        return
    config = get_config(values['config_fn'])
    for wit in values['excluded']:
        config['witnesses'].append(wit)
        config = remove_from_config(config, 'excluded_witnesses', wit)
    save_config(config, values['config_fn'])
    update_window(window, values)

def delete_selection(window: sg.Window, values: dict):
    if values['excluded'] == []:
        return
    config = get_config(values['config_fn'])
    for wit in values['excluded']:
        config = remove_from_config(config, 'excluded_witnesses', wit)
    save_config(config, values['config_fn'])
    update_window(window, values)

def start_ce(values):
    if values['config_fn'] == '':
        return
    cwd = os.getcwd()
    root_dir = Path(values['config_fn']).parent.parent.parent.parent.parent.as_posix()
    print(root_dir)
    os.chdir(root_dir)
    try:
        if platform.system() == 'Windows':
            from subprocess import CREATE_NEW_CONSOLE
            subprocess.Popen('start startup.bat', shell=True, creationflags=CREATE_NEW_CONSOLE)
            subprocess.Popen('start firefox http://localhost:8080/collation', shell=True)
        else:
            subprocess.Popen('bash startup.sh', shell=True)
            webbrowser.get('firefox').open('http:localhost:8080/collation')
    except:
        print('cold not open browser')
    os.chdir(cwd)
    return

def configure_ce(font, icon):
    window = sg.Window('Configure Collation Editor', layout(), font=font, icon=icon, resizable=True)
    while True:
        event, values = window.read()
        if event in [None, sg.WINDOW_CLOSED, 'exit']:
            break
        elif event == 'Update':
            if values['config_fn'] == '':
                continue
            try:
                edit_config(values)
                update_window(window, values)
            except:
                sg.popup_quick_message('Are you sure that is the right config.json file?\nIt did not work.')
            es.edit_settings('ce_config_fn', values['config_fn'])
        elif event == 'Add Witness':
            add_witness(values, window)
        elif event == 'Move to Excluded':
            remove_witnesses(values, window)
        elif event == 'Start Collation Editor':
            start_ce(values)
        elif event == 'Move to Included':
            include_selected(values, window)
        elif event == 'Delete Selected':
            delete_selection(window, values)
    window.close()
    return False



from pathlib import Path

import PySimpleGUI as sg 

import criticus.py.edit_settings as es
from criticus.py.md2tei.markdown_to_tei import convert_md_to_tei as md2tei
import criticus.py.custom_popups as cp

# pylint: disable=no-member

def bummer():
    cp.ok('The Markdown file failed to convert. Talk to David about it.', 'Bummer...')

def yay(output_file):
    msg = f'Yay! The Markdown transcription was converted and saved to\n{output_file}'
    cp.ok(msg, 'Yay!')

def set_initial_dirs(input_file, output_file):
    if input_file:
        initial_dir = Path(input_file).parent.as_posix()
        es.edit_settings('markdown_tx_dir', initial_dir)
    if output_file:
        output_dir = Path(output_file).parent.as_posix()
        es.edit_settings('converted_markdown_tx_dir', output_dir)

def convert(settings, values):
    output_file = sg.popup_get_file('', no_window=True, save_as=True, file_types=(('XML File', '*.xml'), ), initial_folder=settings['converted_markdown_tx_dir'])
    set_initial_dirs(values['md_input'], output_file)
    md2tei(values['md_input'], output_file, values['plain'], values['lines'])
    return output_file

def layout(settings):
    plain_tip = 'smaller file size; not very human-readable'
    lines_tip = 'break lines according to transcription'
    pretty_tip = 'pretty print'
    radio_frame = [
        [sg.Radio('Do not add extra whitespace', 'format', tooltip=plain_tip, key='plain')],
        [sg.Radio('Keep transcription lines', 'format', tooltip=lines_tip, key='lines')],
        [sg.Radio('Pretty Print', 'format', tooltip=pretty_tip, key='pretty', default=True)],
    ]
    io_frame = [
        [sg.I('', key='md_input'), sg.FileBrowse('Browse', file_types=(('Markdown Files', '*.md'),), initial_folder=settings['markdown_tx_dir'])]
    ]
    return [
        [sg.Frame('XML TEI Output Format', radio_frame)],
        [sg.Frame('Markdown File', io_frame)],
        [sg.B('Convert Markdown to TEI', key='convert'), sg.B('Cancel', key='exit')]
    ]

def md_to_tei(font: tuple, icon):
    settings = es.get_settings()
    L = layout(settings)
    window = sg.Window('MarkdownTEI', L, font=font, icon=icon)

    while True:
        event, values = window.read()
        if event in [sg.WINDOW_CLOSED, 'exit']:
            break
        elif event == 'convert' and values['md_input'] != '':
            try:
                output_file = convert(settings, values)
                yay(output_file)
            except:
                bummer()
    window.close()
    return False



import lxml.etree as et
from markdown import Markdown
from criticus.py.md2tei.md_tei_extension import TEI, postprocess_xml, preprocess_md, postprocess_markup # pylint: disable=import-error


def convert_md_to_tei(md_file, xml_file, plain: bool, lines: bool):
    M = Markdown(extensions=['attr_list', TEI(), 'markdown_del_ins'])
    with open(md_file, 'r', encoding='utf-8') as file:
        md = file.read()

    md = preprocess_md(md)
    markup = M.convert(md)
    markup = postprocess_markup(markup)

    parser = et.XMLParser(remove_blank_text=True, encoding='UTF-8')
    xml = et.fromstring(markup, parser)
    xml = postprocess_xml(xml)

    xml = xml.getroottree()
    if plain:
        xml.write(xml_file, encoding='utf-8')
    elif lines:
        xml_str = et.tostring(xml, encoding='unicode')
        xml_str = xml_str.replace('\n', '')
        xml_str = xml_str.replace('<pb', '\n<pb')
        xml_str = xml_str.replace('<lb', '\n    <lb')
        with open(xml_file, 'w', encoding='utf-8') as f:
            f.write(xml_str)
    else:
        et.indent(xml, '    ')
        xml.write(xml_file, encoding='utf-8', pretty_print=True)



import json
from pathlib import Path
from typing import List

def regularize_word(word: str):
    word = word.replace('\u0323', '')
    word = word.replace('[', '')
    word = word.replace(']', '')
    return word

def words_to_tokens(words: list, siglum: str, ):
    tokens = []
    gap_before = False
    gap_after = False
    for i, word in enumerate(words, start=1):
        gap_after = False
        gap_before = False
        if f"___{word.replace('___', '')}" == word:
            gap_before = True
            word = word.replace('___', '')
        elif f"{word.replace('___', '')}___" == word:
            gap_after = True
            word = word.replace('___', '')
        regularized = regularize_word(word)
        token = {'index': f'{i*2}',
                 'siglum': siglum,
                 'reading': siglum,
                 'original': word,
                 'rule_match': [regularized],
                 't': regularized}
        if gap_after:
            token['gap_after'] = True
        if gap_before:
            token['gap_before'] = True
        tokens.append(token)
    return tokens


def dictify_witnesses(witness: tuple, siglum):
    if witness[0] != 'firsthand':
        corrector = witness[0].replace('corrector', 'c')
        siglum = f'{siglum}{corrector}'
    return {
        'id': siglum,
        'tokens': words_to_tokens(witness[1], siglum)
    }

def dictify_transcription(siglum: str, ref: str, plain_tx: str, witnesses: List[dict]) -> dict:
    return {
        'id': f'{siglum}_{ref}',
        'transcription': siglum,
        'transcription_siglum': siglum,
        'siglum': siglum,
        'context': ref, 
        'n': ref,
        'plain_text': plain_tx,
        'witnesses': witnesses
    }

def save_tx(transcription: dict, filename: str, output_dir: str):
    output_path = Path(output_dir)
    output_path.mkdir(exist_ok=True)
    output_dir = output_path.as_posix()
    fn = f'{output_dir}/{filename}.json'
    with open(fn, 'w', encoding='utf-8') as file:
        json.dump(transcription, file, ensure_ascii=False, indent=4)

def verse_to_dict(siglum: str, ref: str, witnesses: List[tuple]) -> dict:
    wits = []
    plain = ' '.join(witnesses[0][1]).replace('___', '')
    for witness in witnesses:
        wits.append(dictify_witnesses(witness, siglum))
    transcription = dictify_transcription(siglum, ref, plain, wits)
    return transcription



import json
from pathlib import Path
import re

from natsort import natsorted
import PySimpleGUI as sg

import criticus.py.edit_settings as es
import criticus.py.custom_popups as cp

# pylint: disable=no-member
def json_to_plain_text(json_dir: str):
    files = Path(json_dir).glob('**/*.json')
    files = [f.as_posix() for f in files]
    files = natsorted(files)
    lines = []
    for verse in files:
        try:
            with open(verse, 'r', encoding='utf-8') as f:
                tx = json.load(f)
            lines.append(f"{tx['context']} {tx['plain_text']}")
        except:
            # print(f'Did not open {verse}')
            pass
    text = '\n'.join(lines)
    return text

def simplify_ref(text: str, icon):
    text = re.sub(r'B[0-9]+K', 'K', text)
    text = text.splitlines()

    chapters = []
    new_lines = []
    skipped_lines = []

    for line in text:
        try:
            ch = re.search(r'K[0-9]+V', line).group(0)
        except:
            # print(f'Did not find a reference on this line:\n{line}')
            skipped_lines.append(line)
            continue
        ch = ch.replace('V', '')
        ch_num = ch.replace('K', '')
        if ch_num not in chapters:
            chapters.append(ch_num)
            new_lines.append(f'\nCHAPTER {ch_num}\n')
        new_lines.append(line)

    text = '\n'.join(new_lines)
    text = re.sub(r'K.+V', '', text)

    if len(skipped_lines) > 0:
        skipped_lines = '\n'.join(skipped_lines)
        cp.ok(f'''The following lines were skipped because a reference was not found:
{skipped_lines}''', title='Some lines skipped')

    return text

def get_folder(font, icon, settings):
    layout = [
        [sg.T('Select a folder containing the target JSON files')],
        [sg.I(settings['txt_from_json_dir'], key='folder'), sg.FolderBrowse(initial_folder=settings['tx_dir'])],
        [sg.B('Get Text'), sg.T(''), sg.B('Cancel')]
    ]
    window = sg.Window('Plain Text from JSON Files', layout, font=font, icon=icon)
    folder = None
    while True:
        event, values = window.read()
        if event in [sg.WINDOW_CLOSED, None, 'Cancel']:
            break
        elif event == 'Get Text':
            folder = Path(values['folder']).absolute().as_posix()
            break
    window.close()
    return folder

def get_text_from_json_files(font, icon):
    settings = es.get_settings()
    folder = get_folder(font, icon, settings)
    if not folder:
        return
    folder = Path(folder).as_posix()
    es.edit_settings('txt_from_json_dir', folder)
    text = json_to_plain_text(folder)
    text = simplify_ref(text, icon)
    save_fn = sg.popup_get_file(
        '', title='Save Plain Text', save_as=True, no_window=True, 
        file_types=(('Plain Text Files', '*.txt'),),
        initial_folder=settings['plain_text_dir'],
        font=font
        )
    if not save_fn:
        return
    save_fn = Path(save_fn).as_posix()
    save_fn_setting = Path(save_fn).parent.as_posix()
    es.edit_settings('plain_text_dir', save_fn_setting)
    with open(save_fn, 'w', encoding='utf-8') as f:
        f.write(text)
    cp.ok(f'Plain text was extracted from JSON files \
and saved to\n{save_fn}', title='Success!')



from pathlib import Path
from types import FunctionType
import platform
import sys

import PySimpleGUI as sg 

from criticus.py.txt2json.window_text_to_json import txt_to_json
from criticus.py.combine_xml import combine_xml_files_interface as combine_xml
from criticus.py.md2tei.MarkdownTEI import md_to_tei
from criticus.py.tei2json.tei2json_ui import tei_to_json
from criticus.py.reformat_collation.reformat_xml_ui import start_reformat_ui as reform
from criticus.py.serve_tei_transcriptions.serve_tei_tx_ui import serve_tei_tx
from criticus.py.ce_config import configure_ce
from criticus.py.txt_from_json import get_text_from_json_files
from criticus.main_layout import main_layout
from criticus.py.export_to_docx.xml_to_docx_ui import export_to_docx
from criticus.py.check_for_updates import check_for_update

# if platform.system() == 'Windows':
from criticus.py.cbgm_interface.open_cbgm_ui import open_cbgm_ui

__version__ = '0.35.2'
#pylint: disable=no-member

def open_new_window(function: FunctionType, window: sg.Window, main_dir, font, icon, include_main_dir=False):
    window.hide()
    stay_open = True
    while stay_open:
        if include_main_dir:
            stay_open = function(main_dir, font, icon)
        else:
            stay_open = function(font, icon)
    window.un_hide()

def get_actual_dir():
    '''Since PyInstaller unpacks modules into a virtual 
    directory, this is required to get the actual absolute 
    path of the main directory both for dev and frozen states.'''
    if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'): # shipped app
        return Path(sys.executable).parent.as_posix()
    else: # dev
        return Path(__file__).parent.as_posix()

def main():
    main_dir = get_actual_dir()
    sg.set_options(dpi_awareness=True)
    sg.LOOK_AND_FEEL_TABLE['Parchment'] = {'BACKGROUND': '#FFE9C6',
                                        'TEXT': '#533516',
                                        'INPUT': '#EAC8A3',
                                        'TEXT_INPUT': '#2F1B0A',
                                        'SCROLL': '#B39B73',
                                        'BUTTON': ('white', '#C55741'),
                                        'PROGRESS': ('#01826B', '#D0D0D0'),
                                        'BORDER': 1, 'SLIDER_DEPTH': 0, 'PROGRESS_DEPTH': 0,
                                        }
    sg.theme('Parchment')
    if platform.system() == 'Windows':
        icon = f'{main_dir}/resources/criticus.ico'
        font = ('Cambria', 11)
    else:
        icon = f'{main_dir}/resources/criticus.png'
        font = ('Arial', 11)
    window = sg.Window(f'Criticus v{__version__}', main_layout(), font=font, icon=icon, debugger_enabled=False)
    while True:
        event, _ = window.read()

        if event in [sg.WIN_CLOSED, None, 'Close']:
            break

        elif event == 'txt_to_json':
            open_new_window(txt_to_json, window, main_dir, font, icon)

        elif event == 'combine_verses':
            open_new_window(combine_xml, window, main_dir, font, icon, include_main_dir=True)

        elif event == 'md_to_tei':
            open_new_window(md_to_tei, window, main_dir, font, icon)

        elif event == 'tei_to_json':
            open_new_window(tei_to_json, window, main_dir, font, icon)

        elif event == 'reformat_xml':
            open_new_window(reform, window, main_dir, font, icon)

        elif event == 'tei_server':
            open_new_window(serve_tei_tx, window, main_dir, font, icon, include_main_dir=True)

        elif event == 'ce_config':
            open_new_window(configure_ce, window, main_dir, font, icon)

        elif event == 'json_to_txt':
            open_new_window(get_text_from_json_files, window, main_dir, font, icon)

        elif event == 'open-cbgm':
            open_new_window(open_cbgm_ui, window, main_dir, font, icon)

        elif event == 'export_to_docx':
            open_new_window(export_to_docx, window, main_dir, font, icon)

        elif event == 'Check for Updates':
            check_for_update(__version__)

    window.close()



import json
from pathlib import Path
import sys

if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'): # shipped app
    main_dir = Path(sys.executable).parent.joinpath('resources').as_posix()
else: # dev
    main_dir = Path(__file__).parent.parent.joinpath('resources').as_posix()


def get_settings():
    try:
        with open(f'{main_dir}/settings.json', 'r') as f:
            settings = json.load(f)
        _ = (
            settings['ce_repo_dir'],
            settings['tx_dir'],
            settings['ce_output_dir'],
            settings['markdown_tx_dir'],
            settings['converted_markdown_tx_dir'],
            settings['tei_dir'],
            settings['combined_xml_dir'],
            settings['reformatted_xml_dir'],
            settings['ce_config_fn'],
            settings['txt_from_json_dir'],
            settings['plain_text_dir'],
            settings['cbgm_main_dir'],
            settings['cbgm_cx_dir'],
            settings['pre_parse_regex'],
            settings['custom_template_path']
            ) 
    except:
        settings = {
            'ce_repo_dir': '',
            'tx_dir': '',
            'ce_output_dir': '',
            'markdown_tx_dir': '',
            'converted_markdown_tx_dir': '',
            'tei_dir': '',
            'combined_xml_dir': '',
            'reformatted_xml_dir': '',
            'ce_config_fn': '',
            'txt_from_json_dir': '',
            'plain_text_dir': '',
            'cbgm_main_dir': '',
            'cbgm_cx_dir': '',
            'pre_parse_regex': [],
            'export_docx_folder': '',
            'text_wits_separator': ' // ',
            'rdg_n_text_separator': '\t',
            'words_per_line': 10,
            'text_bold': False,
            'custom_template_path': ''
        }
        save_settings(settings)
    return settings

def save_settings(settings: dict):
    with open(f'{main_dir}/settings.json', 'w') as f:
        json.dump(settings, f, indent=4)

def edit_settings(setting_key: str, value):
    settings = get_settings()
    settings[setting_key] = value
    save_settings(settings)



from pathlib import Path
import platform
from turtle import back

import PySimpleGUI as sg

# pylint: disable=no-member
op_os = platform.system()

def get_icon():
    criticus_dir = Path(__file__).parent.parent.as_posix()
    return f'{criticus_dir}/resources/criticus.ico'

def ok(msg, title):
    layout = [[sg.T(msg)],
              [sg.B('Ok')]]
    window = sg.Window(title, layout, icon=get_icon())
    window.read()
    window.close()

def yes_cancel(msg, title):
    layout = [[sg.T(msg)],
              [sg.B('Yes'), sg.T(''), sg.B('Cancel')]]
    window = sg.Window(title, layout, icon=get_icon())
    event, _ = window.read()
    window.close()
    if event == 'Yes':
        return True
    else:
        return False

def textbox(text, title):
    if op_os == 'Windows':
        size = (1200, 400)
    else:
        size = (800, 400)
    layout = [
        [sg.Multiline(text, font=('Courier', 10), expand_x=True, expand_y=True)],
        [sg.B('Done')]
        ]
    window = sg.Window(title, layout, icon=get_icon(), resizable=True, size=size)
    window.read()
    window.close()

def listbox(message: str, items: list, title: str):
    layout = [[sg.T(message)],
              [sg.Listbox(items)],
              [sg.B('Ok')]]
    window = sg.Window(title, layout, icon=get_icon())
    window.read()
    window.close()

def mac_win_cmd(text: str, title: str, mac_cmd: str, win_cmd: str):
    layout = [
        [sg.T(text)],
        [sg.T('MacOS:', size=(10, 1)), sg.Input(mac_cmd, readonly=True, disabled_readonly_background_color='#000000', disabled_readonly_text_color='#ffffff')],
        [sg.T('Windows:', size=(10, 1)), sg.Input(win_cmd, readonly=True, disabled_readonly_background_color='#000000', disabled_readonly_text_color='#ffffff')],
        [sg.Stretch(), sg.Button('Ok', size=(5, 1)), sg.Stretch()],
    ]
    window = sg.Window(title, layout, icon=get_icon())
    window.read()
    window.close()
    


import codecs
import re
from typing import List

import lxml.etree as et

import criticus.py.edit_settings as es
import criticus.py.custom_popups as ce

###########################################################
###########################################################
# namespaces
tei_ns = 'http://www.tei-c.org/ns/1.0'
xml_ns = 'http://www.w3.org/XML/1998/namespace'

def get_file(tei: str):
    with open(tei, 'r', encoding='utf-8') as f:
        return f.read()

###########################################################
###########################################################

# pre parse cleanup
def unescape_string(text: str):
    cleaned = codecs.decode(text, 'unicode-escape')
    return cleaned

def pre_parse_cleanup(text):
    settings = es.get_settings()
    text = re.sub(r' +|\t', ' ', text)
    text = re.sub(r' *<supplied[^<>]*/>', '', text)
    text = re.sub(r' *<supplied[^<>]*>', '[', text)
    text = re.sub(r' *</supplied> *', ']', text)
    text = re.sub(r'<lb[^<>]*>', '', text)
    text = re.sub(r'</lb[^<>]*>', '', text)
    text = re.sub(r'<cb[^<>]*>', '', text)
    text = re.sub(r'</cb[^<>]*>', '', text)
    text = re.sub(r'<pb[^<>]*>', '', text)
    text = re.sub(r'</pb[^<>]*>', '', text)
    text = re.sub(r'<abbr[^<>]*>', '', text)
    text = re.sub(r'</abbr[^<>]*>', '', text)
    text = re.sub(r'<hi[^<>]*>', '', text)
    text = re.sub(r'</hi[^<>]*>', '', text)
    text = re.sub(r'<ex[^<>]*>', '', text)
    text = re.sub(r'</ex>', '', text)
    text = re.sub(r'<pc>[^<>]+</pc>', '', text)
    # text = re.sub(r'(<rdg[^<>]*>)<seg[^<>]*>', r'\1', text)
    # text = re.sub(r'</seg></rdg>', '</rdg>', text)
    text = text.replace('\n', '')
    text = text.replace('encoding="utf-8"', '')
    for regex in settings['pre_parse_regex']:
        try:
            to_replace = unescape_string(regex[0])
            replacement = unescape_string(regex[1])
            text = re.sub(to_replace, replacement, text)
            print(f'{to_replace=}\n{replacement=}')
        except Exception as e:
            ce.ok(f'Failed regular expression: {regex}\n{e}', 'RegEx Fail')
    return text

def parse(text: str):
    parser = et.XMLParser(remove_blank_text=True, encoding='utf-8', recover=True)
    try:
        return (True, et.fromstring(text, parser))
    except Exception as e:
        return (False, e)


###########################################################
###########################################################
# after parsed

def add_underdot_to_unclear_letters(root: et._Element): #* PASSING
    unclear = root.xpath('//tei:unclear', namespaces={'tei': tei_ns})
    for u in unclear:
        underdotted = []
        if u.text is None:
            print(f'NONE WORD!!!{et.tostring(u, encoding="unicode")=}')
        for letter in u.text:
            underdotted.append(f'{letter}\u0323')
        u.text = ''.join(underdotted)

def remove_unclear_tags(text: str): #* PASSING
    text = re.sub(r'<unclear[^<>]*>', '', text)
    text = re.sub(r'</unclear[^<>]*>', '', text)
    return text

def write_xml(root):
    tree = root.getroottree()
    tree.write('output.xml', encoding='utf-8', pretty_print=True)

def handle_abbr(abbr: et._Element): #* PASSING
    if not abbr.text and abbr.getchildren():
        if abbr.getchildren()[0].text:
            return abbr.getchildren()[0].text
    elif abbr.text and not abbr.getchildren():
        return abbr.text

def handle_app(app: et._Element, hand: str):
    def get_text(word: et._Element, words: list):
        if word.text and word.tag != '{http://www.tei-c.org/ns/1.0}pc':
            words.append(word.text)
        elif word.getchildren():
            if word.getchildren()[0].text and word.tag != '{http://www.tei-c.org/ns/1.0}pc':
                words.append(word.getchildren()[0].text)
            elif word.getchildren()[0].tag == '{http://www.tei-c.org/ns/1.0}abbr':
                t = handle_abbr(word.getchildren()[0])
                if t:
                    words.append(t)
        return words
    words = []
    if hand == 'firsthand':
        rdg_type = 'orig'
    else:
        rdg_type = 'corr'
    for rdg in app.getchildren():
        if rdg.get('type') == rdg_type and rdg.get('hand') == hand:
            for elem in rdg.getchildren():
                if elem.tag == '{http://www.tei-c.org/ns/1.0}seg':
                    for word in elem.getchildren():
                        words = get_text(word, words)
                else:
                    words = get_text(elem, words)
    return words

def get_all_words_in_verse(verse: et._Element, hand: str) -> List[str]:
    words = []
    for elem in verse:
        if elem.tag == f'{{{tei_ns}}}w' and not elem.getchildren() and elem.text:
            words.append(elem.text)
        elif elem.tag == f'{{{tei_ns}}}w' and elem.getchildren():
            for child in elem:
                if child.tag == f'{{{tei_ns}}}abbr':
                    words.append(handle_abbr(child))
                elif child.text:
                    words.append(child.text)
        elif elem.tag == f'{{{tei_ns}}}app':
            words += handle_app(elem, hand)  
    return words

def handle_lacunae(words: List[str]) -> List[str]: ###* PASSING
    indices_of_lac_words = []
    for i, word in enumerate(words):
        if not word:
            print(f'\n{word=}\n')
        lac_word_found = re.search(r'\[[^\[\]]*\]', word)
        if lac_word_found and lac_word_found.group(0) == word: # entire word is supplied, i.e. lacunose b/c word == [word]
            indices_of_lac_words.append(i)
    if 0 in indices_of_lac_words and 1 not in indices_of_lac_words:
        words[1] = words[1].replace(words[1], f'___{words[1]}')
    furthest_lac_index = None
    for index in indices_of_lac_words:
        if furthest_lac_index:
            if index <= furthest_lac_index:
                continue
            else:
                furthest_lac_index = None
        if index + 1 not in indices_of_lac_words and index != 0:
            words[index-1] = words[index-1].replace(words[index-1], f'{words[index-1]}___')
        elif index + 1 in indices_of_lac_words:
            words[index-1] = words[index-1].replace(words[index-1], f'{words[index-1]}___')
            for i, _ in enumerate(words, 1):
                if index + i in indices_of_lac_words:
                    furthest_lac_index = index + i
                else:
                    break 
    new_words = []
    for i, w in enumerate(words):
        if i in indices_of_lac_words:
            continue
        else:
           new_words.append(w) 
    return new_words

def remove_duplicate_witnesses(witnesses: List[list]):
    found_wits = []
    new_wits = []
    for w in witnesses:
        if w[1] not in found_wits:
            found_wits.append(w[1])
            new_wits.append(w)
    return new_wits

def get_verse_as_tuple(verse: et._Element, hands: list = ['firsthand']) -> List[tuple]:
    witnesses = []
    for hand in hands:
        words = get_all_words_in_verse(verse, hand)
        words = handle_lacunae(words)
        witnesses.append((hand, words))
    witnesses = remove_duplicate_witnesses(witnesses)
    return witnesses



import json
from pathlib import Path

import platform
operating_system = platform.system()

if operating_system == 'Windows':
    from subprocess import Popen, CREATE_NEW_CONSOLE, check_output
else:
    from subprocess import Popen, check_output
import os

# pylint: disable=import-error
import criticus.py.custom_popups as cp
import criticus.py.edit_settings as es

def get_system_specific_command(command, new_console: bool):
    if operating_system == 'Windows':
        if new_console:
            p = Popen(command, creationflags=CREATE_NEW_CONSOLE)
        else:
            p = Popen(command)
    else:
        command = command.replace(".exe", "")
        command = command.replace('\\', '/')
        p = Popen(command, shell=True)
    return p

def get_system_output_command(command):
    if operating_system == 'Windows':
        return command
    command = command.replace('\\', '/')
    command = command.replace(".exe", "")
    return command

def parse_user_input(values: dict):
    settings = es.get_settings()
    db = Path(f'{settings["cbgm_main_dir"]}/db/{values["new_db_name"]}.db').as_posix().replace('/', '\\')
    cx = Path(values['xml_file']).as_posix().replace('/', '\\')
    script = Path(f"{settings['cbgm_main_dir']}/populate_db.exe").as_posix().replace('/', '\\')
    command = f'"{script}"'
    if values['threshold']:
        for threshold in values['threshold_input'].strip().split(', '):
            command = f'{command} -t {threshold}'
    if values['trivial']:
        for t in values['trivial_input'].strip().split(', '):
            command = f'{command} -z {t}'
    if values['exclude']:
        for e in values['exclude_input'].strip().split(', '):
            command = f'{command} -Z {e}'
    if values['merge_split']:
        command = f'{command} --merge-splits'
    if values['classic']:
        command = f'{command} --classic'
    return f'{command} "{cx}" "{db}"'

def check_db_dir(cbgm_main_dir: str):
    db_dir = Path(f'{cbgm_main_dir}/db')
    if not db_dir.is_dir():
        db_dir.mkdir()
    if db_dir.is_dir():
        return True
    else:
        return False

def populate_db(values: dict):
    if not check_db_dir(values['cbgm_main_dir']):
        cp.ok('There was an unanticipated problem making a folder\n\
in which to save databases.')
        return
    command = parse_user_input(values)
    p = get_system_specific_command(command, new_console=True)
    p.wait()

def get_all_dbs():
    settings = es.get_settings()
    try:
        db_dir = f"{settings['cbgm_main_dir']}/db"
        db_dir = Path(db_dir)
        databases = [x.as_posix().split('/')[-1] for x in db_dir.iterdir() if x.as_posix().endswith('.db')]
        if databases is None or databases == []:
            databases = ['']
    except:
        databases = ['']
    return databases

def delete_db(values):
    settings = es.get_settings()
    db_dir = f"{settings['cbgm_main_dir']}/db"
    db_dir = Path(db_dir)
    for db in values['db_listbox']:
        for f in db_dir.iterdir():
            if f.as_posix().endswith(db):
                try:
                    os.remove(f.as_posix())
                except Exception as e:
                    # print(f'HEADS UP: {e}')
                    cp.ok(f'There was a problem deleting\n{f.as_posix()}\nbecause: {e}', 'Just a small problem')

def parse_compare_input(values):
    settings = es.get_settings()
    script = Path(f'''{settings['cbgm_main_dir']}/compare_witnesses.exe''').as_posix().replace('/', '\\')
    db = Path(f'{settings["cbgm_main_dir"]}/db/{values["selected_db"]}').as_posix().replace('/', '\\')
    command = f'"{script}" -f json "{db}" {values["wit_to_compare"]}'
    for wit in values['wits_to_compare'].strip().split(', '):
        command = f'{command} {wit}'
    return command

def compare_wits(values):
    command = parse_compare_input(values)
    command = get_system_output_command(command)
    try:
        text = check_output(command, shell=True)
    except:
        return None
    text = text.decode()
    text = text.replace('Opening database...', '')
    text = text.replace('Retrieving witness list...', '')
    text = text.replace('Retrieving genealogical relationships for primary witness...', '')
    text = text.replace('Closing database...', '')
    text = text.replace('Database closed.', '')
    text = text.replace('Writing to cout...', '')
    text = text.strip()
    return json.loads(text)

def csv_comparison(values, output_fn):
    output_fn = Path(output_fn).as_posix().replace('/', '\\')
    settings = es.get_settings()
    script = Path(f'''{settings['cbgm_main_dir']}/compare_witnesses.exe''').as_posix().replace('/', '\\')
    db = Path(f'{settings["cbgm_main_dir"]}/db/{values["selected_db"]}').as_posix().replace('/', '\\')
    command = f'"{script}" -f csv -o "{output_fn}" "{db}" {values["wit_to_compare"]}'
    for wit in values['wits_to_compare'].strip().split(', '):
        command = f'{command} {wit}'
    p = get_system_specific_command(command, new_console=False)
    p.wait()
    if Path(output_fn).is_file():
        return True
    else:
        return False

def view_plain_text(values):
    settings = es.get_settings()
    script = Path(f'''{settings['cbgm_main_dir']}/compare_witnesses.exe''').as_posix().replace('/', '\\')
    db = Path(f'{settings["cbgm_main_dir"]}/db/{values["selected_db"]}').as_posix().replace('/', '\\')
    command = f'"{script}" "{db}" {values["wit_to_compare"]}'
    for wit in values['wits_to_compare'].strip().split(', '):
        command = f'{command} {wit}'
    command = get_system_output_command(command)
    try:
        text = check_output(command, shell=True)
    except:
        text = b'Operation Failed'
    text = text.decode()
    text = text.replace('Opening database...', '')
    text = text.replace('Retrieving witness list...', '')
    text = text.replace('Retrieving genealogical relationships for primary witness...', '')
    text = text.replace('Closing database...', '')
    text = text.replace('Database closed.', '')
    text = text.replace('Writing to cout...', '')
    text = text.strip()
    return text



import json
from pathlib import Path
from typing import List
import criticus.py.edit_settings as es


def get_file(filename):
    with open(filename, 'r', encoding='utf-8') as f:
        text = f.readlines()
    return text

def get_info_from_filename(filename):
    filename = filename.split('/')[-1]
    f = filename.split('_')
    siglum = f[0]
    reference_prefix = f[1].replace('.txt', '')
    return siglum, reference_prefix

def format_reference(line: List[str], reference_prefix: str):
    if reference_prefix[-1].isdigit():
        seperator = '.'
    else:
        seperator = ''
    verse = line.pop(0)
    reference = verse.replace(':', '.')
    reference = f'{reference_prefix}{seperator}{reference}'
    return line, reference, verse

def convert_this_line(verse: str, verse_from, verse_to) -> bool:
    if verse.replace(':', '.') in [verse_from.replace(':', '.'), verse_to.replace(':', '.')]:
        return True
    else:
        return False

def build_token(word: str, index: str, siglum: str) -> dict:
    return {
        'index': index,
        'siglum': siglum,
        'reading': siglum,
        'original': word,
        'rule_match': [word],
        't': word
    }

def build_witneses(siglum: str, tokens: List[dict]):
    return [{
        'id': siglum,
        'tokens': tokens
    }]

def build_json(siglum, reference, witnesses: List[dict], line: list):
    return {
        'id': siglum,
        '_id': siglum,
        'transcription': siglum,
        'transcription_siglum': siglum,
        'siglum': siglum,
        'context': reference,
        'n': reference,
        'text': ' '.join(line),
        'witnesses': witnesses
    }

def make_tokens(line, siglum) -> List[dict]:
    tokens = []
    for i, word in enumerate(line, 1):
        index = f'{i*2}'
        token = build_token(word, index, siglum)
        tokens.append(token)
    return tokens

def save_json(to_save: dict, reference: str, output_dir: str):
    with open(f'{output_dir}/{reference}.json', 'w', encoding='utf-8') as f:
        json.dump(to_save, f, ensure_ascii=False, indent=4)

def save_metadata(siglum, output_dir):
    metadata = {'_id': siglum, 'siglum': siglum, 'id': siglum}
    with open(f'{output_dir}/metadata.json', 'w', encoding='utf-8') as f:
        json.dump(metadata, f, ensure_ascii=False)

def construct_json_transcription(line: List[str], reference: str, siglum: str, output_dir):
    tokens = make_tokens(line, siglum)
    witnesses = build_witneses(siglum, tokens)
    complete_json = build_json(siglum, reference, witnesses, line)
    save_json(complete_json, reference, output_dir)

def check_and_save_dirs(output_dir, siglum):
    output_dir = Path(f'{output_dir}/{siglum}')
    if not output_dir.exists():
        Path.mkdir(output_dir, parents=True)
    es.edit_settings('output_dir', output_dir.parent.absolute().as_posix())
    return output_dir.absolute().as_posix()

def convert_text_to_json(
    filename, output_dir, convert_all: bool,
    reference_prefix: str, auto: bool, verse_from: str=None, 
    verse_to: str=None, siglum: str=None
    ):
    filename = Path(filename).as_posix()
    if auto:
        siglum, reference_prefix = get_info_from_filename(filename)
    output_dir = check_and_save_dirs(output_dir, siglum)
    save_metadata(siglum, output_dir)  
    text = get_file(filename)
    capture = False
    for line in text:
        line = line.split()
        if len(line) > 1 and line[0][0].isdigit() and line[0][-1].isdigit(): # check that line contains a reference and a text unit and is not a heading
            line, reference, verse = format_reference(line, reference_prefix)

            if not convert_all and verse_to != verse_from: # handle a range of text units to convert
                if convert_this_line(verse, verse_from, verse_to):
                    capture = not capture
                    construct_json_transcription(line, reference, siglum, output_dir)
                elif capture:
                    construct_json_transcription(line, reference, siglum, output_dir)

            elif not convert_all and verse_from == verse_to: # handle a single text unit
                if convert_this_line(verse, verse_from, verse_to):
                    construct_json_transcription(line, reference, siglum, output_dir)

            else:
                construct_json_transcription(line, reference, siglum, output_dir) # handle all text units



import PySimpleGUI as sg 
import criticus.py.edit_settings as es
from pathlib import Path
import os
import threading

from criticus.py.serve_tei_transcriptions.serve_tei import launch_tei_viewer as launch

# pylint: disable=no-member
def layout(settings: dict):
    tei_repo_frame = [
        [sg.I(settings['tei_dir'], key='tei_folder'), 
         sg.FolderBrowse(initial_folder=settings['tei_dir'])]
    ]
    return [
        [sg.Frame('TEI Transcription Folder', tei_repo_frame)],
        [sg.B('Launch TEI Transcription Viewer', key='launch'), sg.B('Cancel', key='exit')]
    ]

def set_initial_dir(tei_dir: str):
    tei_dir = Path(tei_dir).as_posix()
    es.edit_settings('tei_dir', tei_dir)

def kill_server(httpd):
    httpd.shutdown()
    print('server stopped')

def serve_tei_tx(main_dir, font: tuple, icon):
    cwd = Path.cwd().as_posix()
    settings = es.get_settings()
    window = sg.Window('Launch TEI Transcription Viewer', layout(settings), font=font, icon=icon)
    httpd = None
    while True:
        event, values = window.read()

        if event in ['exit', sg.WINDOW_CLOSED, None]:
            if httpd:
                threading.Thread(target=kill_server, args=[httpd]).start()
            break

        elif event == 'launch':
            if values['tei_folder'] == '':
                continue
            try:
                tei_dir = Path(values['tei_folder'])
            except:
                continue
            if tei_dir.is_file():
                tei_dir = tei_dir.parent.as_posix()
            else:
                tei_dir = tei_dir.as_posix()
            set_initial_dir(tei_dir)
            httpd = launch(tei_dir, main_dir)
            window['launch'].update(disabled=True)
    
    window.close()
    os.chdir(cwd)
    return False


"""
Many thanks to Joey McCollum, from whose code this was slightly adapted. 
(https://github.com/jjmccollum/itsee-to-open-cbgm)
It was downloaded on June 1, 2020

The only modification I have made occurs in the main function:
This script is now a module that is called from a main application 
instead of a standalone command line utility.
"""

"""MIT License

Copyright (c) 2020 Joey McCollum

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
from lxml import etree as et

"""
XML namespaces
"""
xml_ns = 'http://www.w3.org/XML/1998/namespace'
tei_ns = 'http://www.tei-c.org/ns/1.0'

"""
Book index to SBL book abbreviation Dictionary
"""
books_by_n = {
    'B01': 'Matt',
    'B02': 'Mark',
    'B03': 'Luke',
    'B04': 'John', 
    'B05': 'Acts',
    'B06': 'Rom',
    'B07': '1 Cor',
    'B08': '2 Cor', 
    'B09': 'Gal',
    'B10': 'Eph',
    'B11': 'Phil',
    'B12': 'Col',
    'B13': '1 Thess',
    'B14': '2 Thess',
    'B15': '1 Tim',
    'B16': '2 Tim',
    'B17': 'Titus',
    'B18': 'Phlm',
    'B19': 'Heb',
    'B20': 'Jas', 
    'B21': '1 Pet',
    'B22': '2 Pet', 
    'B23': '1 John',
    'B24': '2 John', 
    'B25': '3 John', 
    'B26': 'Jude',
    'B27': 'Rev'
}

"""
Returns a List of all witness IDs encountered in the input XML tree.
"""
def get_wits(xml):
    wits = []
    distinct_wits = set()
    for rdg in xml.xpath('//tei:rdg', namespaces={'tei': tei_ns}):
        if not rdg.get('wit'):
            ar = rdg.getnext()
            rdg.getparent().remove(rdg)
            ar.attrib['n'] = 'a'
            ar.attrib.pop('type')
            continue
        for wit in rdg.get('wit').split():
            if wit not in distinct_wits:
                distinct_wits.add(wit)
                wits.append(wit)
    return wits

"""
Adds a <teiHeader> with appropriate subelements as the first child of the input XML's <TEI> element.
The <listWit> element in particular is needed for the open-cbgm library.
"""
def add_tei_header(xml, title_stmt: str, publication_stmt: str):
    #Get a List of witness sigla first:
    wits = get_wits(xml)
    #Get the <TEI> element:
    TEI = xml.xpath('//tei:TEI', namespaces={'tei': tei_ns})[0]
    #Append a <teiHeader> element to it:
    teiHeader = et.Element('teiHeader', nsmap={None: tei_ns, 'xml': xml_ns})
    TEI.insert(0, teiHeader)
    #Append a <fileDesc> element to the teiHeader:
    fileDesc = et.Element('fileDesc', nsmap={None: tei_ns, 'xml': xml_ns})
    teiHeader.append(fileDesc)
    #Append a <titleStmt> element to the fileDesc:
    titleStmt = et.Element('titleStmt', nsmap={None: tei_ns, 'xml': xml_ns})
    fileDesc.append(titleStmt)
    #Append a placeholder paragraph to the titleStmt:
    titleStmt_p = et.Element('p', nsmap={None: tei_ns, 'xml': xml_ns})
    # TODO: Replace these 'temporary' statements with user-supplied statements
    titleStmt_p.text = title_stmt
    titleStmt.append(titleStmt_p)
    #Append a <publicationStmt> element to the fileDesc:
    publicationStmt= et.Element('titleStmt', nsmap={None: tei_ns, 'xml': xml_ns})
    fileDesc.append(publicationStmt)
    #Append a placeholder paragraph to the publicationStmt:
    publicationStmt_p = et.Element('p', nsmap={None: tei_ns, 'xml': xml_ns})
    publicationStmt_p.text = publication_stmt
    publicationStmt.append(publicationStmt_p )
    #Append a <sourceDesc> element to the fileDesc:
    sourceDesc = et.Element('sourceDesc', nsmap={None: tei_ns, 'xml': xml_ns})
    fileDesc.append(sourceDesc)
    #Append a <listWit> element to the sourceDesc:
    listWit = et.Element('listWit', nsmap={None: tei_ns, 'xml': xml_ns})
    sourceDesc.append(listWit)
    #Append a <witness> element for each witness we have to the listWit:
    for wit in wits:
        witness = et.Element('witness', nsmap={None: tei_ns, 'xml': xml_ns})
        witness.set('n', wit)
        listWit.append(witness)
    return

"""
Strips the input XML of all <wit> elements.
(These elements are not needed, as the "wit" attribute of their parent <rdg> element 
contains the same information.)
"""
def strip_wit_subelements(xml):
    for wit in xml.xpath('//tei:wit', namespaces={'tei': tei_ns}):
        wit.getparent().remove(wit)
    return

"""
Strips the input XML of every <app> element without "from" and "to" elements 
(i.e, superfluous apparatus elements used to describe which witnesses are lacunose for entire verses).
"""
def strip_unitless_apps(xml):
    for app in xml.xpath('//tei:app[not(@from) and not(@to)]', namespaces={'tei': tei_ns}):
        app.getparent().remove(app)
    return

"""
Converts escaped Unicode character codes for combining underdots to the corresponding characters.
"""
def unescape_underdots(xml):
    for rdg in xml.xpath('//tei:lem|//tei:rdg', namespaces={'tei': tei_ns}):
        if rdg.text is not None and '&amp;#803;' in rdg.text:
            rdg.text = rdg.text.replace('&amp;#803;', '\u0323')
    return

"""
Strips the text from all <lem> and <rdg> elements in the input XML that have type="om".
"""
def strip_om_text(xml):
    for rdg in xml.xpath('//tei:lem[@type="om"]|//tei:rdg[@type="om"]', namespaces={'tei': tei_ns}):
        rdg.text = None
    return

"""
Replaces <app> elements having only one reading in the input XML with <seg> elements containing that reading.
"""
def sub_segs_for_apps(xml):
    for app in xml.xpath('//tei:app', namespaces={'tei': tei_ns}):
        rdgs = app.xpath('tei:rdg', namespaces={'tei': tei_ns})
        if len(rdgs) == 1:
            seg = et.Element('seg', nsmap={None: tei_ns, 'xml': xml_ns})
            seg.text = rdgs[0].text
            app.getparent().replace(app, seg)
    return

"""
Under each <app> element in the input XML, 
adds a <note> element containing a variation unit label, 
a default connectivity value, and a local stemma without edges.
"""
def add_app_notes(xml):
    for app in xml.xpath('//tei:app', namespaces={'tei': tei_ns}):
        #First, get the "n", "from", and "to" attributes of this apparatus:
        app_n = app.get('n')
        app_from = app.get('from')
        app_to = app.get('to')
        #Then get a List of its reading numbers:
        rdg_ids = []
        for rdg in app.xpath('tei:rdg', namespaces={'tei': tei_ns}):
            if rdg.get('n') is not None:
                rdg_ids.append(rdg.get('n'))
        #Add a <note> element as the last child of the <app> element:
        note = et.Element('note', nsmap={None: tei_ns, 'xml': xml_ns})
        app.append(note)
        #Then add a <label> element under the note, if one can be constructed from the attributes of the apparatus:
        match = re.match('(B\d+)(K\d+)(V\d+)', app_n)
        if match is not None:
            book_n = match.groups()[0]
            chapter_n = match.groups()[1]
            verse_n = match.groups()[2]
            label_text = books_by_n[book_n] + ' ' + chapter_n[1:] + ':' + verse_n[1:] 
            if app_from is not None:
                label_text += '/' + app_from
            if app_to is not None and app_to != app_from:
                label_text += '-' + app_to
            label = et.Element('label', nsmap={None: tei_ns, 'xml': xml_ns})
            label.text = label_text
            note.append(label)
        #Then add a <fs> element under the note:
        fs = et.Element('fs', nsmap={None: tei_ns, 'xml': xml_ns})
        note.append(fs)
        #Then add an <f> element for the "connectivity" feature under the feature set:
        f = et.Element('f', nsmap={None: tei_ns, 'xml': xml_ns})
        f.set('name', 'connectivity')
        fs.append(f)
        #Then add a <numeric> element under the feature with the default connectivity value:
        numeric = et.Element('numeric', nsmap={None: tei_ns, 'xml': xml_ns})
        numeric.set('value', '10')
        f.append(numeric)
        #Then add a <graph> element under the note:
        graph = et.Element('graph', nsmap={None: tei_ns, 'xml': xml_ns})
        graph.set('type', 'directed')
        note.append(graph)
        #Then, add a <node> element for each reading under the graph:
        for rdg_id in rdg_ids:
            node = et.Element('node', nsmap={None: tei_ns, 'xml': xml_ns})
            node.set('n', rdg_id)
            graph.append(node)
    return


def reformat_xml(input_addr, title_stmt, publication_stmt):
    output_addr = 'temp_xml_collation_file'
    parser = et.XMLParser(remove_blank_text=True)
    xml = et.parse(input_addr, parser)
    add_tei_header(xml, title_stmt, publication_stmt)
    strip_wit_subelements(xml)
    strip_unitless_apps(xml)
    strip_om_text(xml)
    sub_segs_for_apps(xml)
    add_app_notes(xml)
    xml.write(output_addr, encoding='utf-8', xml_declaration=True, pretty_print=True)
    return output_addr
    


from pathlib import Path

from lxml import etree as et
import PySimpleGUI as sg

from criticus.py.reformat_collation.itsee_to_open_cbgm import reformat_xml
from criticus.py.reformat_collation.differentiate_subreading_ids import differentiate_subreading_ids as diff_ids
from criticus.py.reformat_collation.clean_wits import remove_redundant_correctors
import criticus.py.edit_settings as es
import criticus.py.custom_popups as cp

# pylint: disable=no-member
def layout(settings: dict):
    titlestmt_tip = '''E.g. "Collation of Romans with Emphasis on Lectionary Witnesses"'''
    pub_tip = '''E.g. "This is an open access work licensed under a Creative Commons Attribution 4.0 International license."'''
    input_frame = [
        [sg.I('', key='xml_input_file'), sg.FileBrowse(initial_folder=settings['combined_xml_dir'], file_types=(('XML Files', '*.xml'), ))]
    ]
    return [
        [sg.Frame('Combined Collation File', input_frame)],
        [sg.T('Collation Title:'), sg.Input('untitled', key='title_stmt', expand_x=True, tooltip=titlestmt_tip)],
        [sg.T('Publication Statement:'), sg.Input('unspecified', key='publication_stmt', expand_x=True, tooltip=pub_tip)],
        [sg.B('Convert', key='convert'), sg.B('Cancel', key='exit'), sg.Stretch(), sg.B('Remove Redundant Correctors', k='clean_wits')],
        ]

def set_initial_dirs(combined_xml_dir: str, reformatted_xml_dir: str):
    combined_xml_dir = Path(combined_xml_dir).parent.as_posix()
    reformatted_xml_dir = Path(reformatted_xml_dir).parent.as_posix()
    es.edit_settings('combined_xml_dir', combined_xml_dir)
    es.edit_settings('reformatted_xml_dir', reformatted_xml_dir)

def fix_NCNames(xml_fn):
    '''lxml throws a syntax eror when the id attribute is set to a number, even though it is a string.
    The Collation Editor is allowing invalid XML to be created, I guess'''
    fixed = 'temp_repaired_xml'
    with open(xml_fn, 'r', encoding='utf-8') as f:
        xml = f.read()
    xml = xml.replace('xml:id="1', 'xml:id="I')
    xml = xml.replace('xml:id="2', 'xml:id="II')
    xml = xml.replace('xml:id="3', 'xml:id="III')
    with open(fixed, 'w', encoding='utf-8') as f:
        f.write(xml)
    return fixed

def convert(values, settings):
    if values['xml_input_file'] == '':
        sg.popup_quick_message('Browse to select an XML collation file to convert')
        return
    try:
        try:
            xml_fn = reformat_xml(values['xml_input_file'], values['title_stmt'], values['publication_stmt'])
        except et.XMLSyntaxError:
            sg.popup_quick_message('file contains invalid NCNames, attempting to repair...')
            xml_fn = fix_NCNames(values['xml_input_file'])
            try:
                xml_fn = reformat_xml(xml_fn, values['title_stmt'], values['publication_stmt'])
            except Exception as e:
                cp.ok(f'There are invalid elements or attributes in this XML file.\nError: {e}',
                'Failed to Reformat')
                return
        xml = diff_ids(xml_fn)
        cp.ok('XML Collation File was successfully reformatted!\n\
You will now be prompted to save the converted file.', title='Success!')
        fn_to_save = sg.popup_get_file('', no_window=True, save_as=True, file_types=(('XML Files', '*.xml'),), initial_folder=settings['reformatted_xml_dir'])
        if not fn_to_save:
            return
        xml.write(fn_to_save, encoding='utf-8', xml_declaration=True, pretty_print=True)
    except Exception as e:
        cp.ok(f'Failed to reformat XML file.\n\
Check that the input file name is correct and that\n\
this is the output of the ITSEE Collation Editor.\n\
Error: {e}', title='Bummer...')
        return
    set_initial_dirs(values['xml_input_file'], fn_to_save)


def clean_wits(values: dict, settings: dict):
    if values['xml_input_file'] == '':
        sg.popup_quick_message('Browse to select an XML collation file to convert')
        return
    try:
        xml = remove_redundant_correctors(values['xml_input_file'])
    except Exception as e:
        cp.ok(f'Failed to clean wits.\n{e}', 'Bummer')
        return
    cp.ok('Witnesses in the collation file have been saved, ready to save.', 'Cleaned!')
    fn_to_save = sg.popup_get_file('', no_window=True, save_as=True, file_types=(('XML Files', '*.xml'),), initial_folder=settings['reformatted_xml_dir'])
    if not fn_to_save:
        return
    try:
        xml.write(fn_to_save, encoding='utf-8', xml_declaration=True, pretty_print=True)
        cp.ok(f'Collation was saved to\n{fn_to_save}', 'Saved!')
    except Exception as e:
        cp.ok(f'Could not write the file to disk.\n{e}', 'Bummer')
        return
    set_initial_dirs(values['xml_input_file'], fn_to_save)

def start_reformat_ui(font: tuple, icon):
    settings = es.get_settings()
    window = sg.Window('Reformat XML Collation File', layout(settings), font=font, icon=icon)

    while True:
        event, values = window.read()

        if event in ['exit', sg.WINDOW_CLOSED, None]:
            break

        elif event == 'convert':
            convert(values, settings)

        elif event == 'clean_wits':
            clean_wits(values, settings)

    window.close()
    return False


import re

from lxml import etree as et


def clean_wits(wits) -> str:
    cleaned_wits = list(wits)
    for wit in wits:
        if 'c' not in wit:
            continue
        first_hand = re.sub(r'c1|c2|c3|c4|c5|c', '', wit)
        if first_hand in wits:
            cleaned_wits.remove(wit)
    return ' '.join(cleaned_wits)


def remove_redundant_correctors(input_addr):
    tei_ns = 'http://www.tei-c.org/ns/1.0'

    parser = et.XMLParser(remove_blank_text=True)
    xml = et.parse(input_addr, parser)
    for rdg in xml.xpath('//tei:rdg', namespaces={'tei': tei_ns}):
        wits = rdg.get('wit').split()
        cleaned_wits = clean_wits(wits)
        rdg.attrib['wit'] = cleaned_wits
    return xml


import PySimpleGUI as sg 



#pylint: disable=no-member
def main_layout():
    bs = (32, 1)

    menu = [['File', ['Check for Updates']]]
    layout = [
        [sg.Menu(menu)],
        [sg.Button('Plain Text to JSON', key='txt_to_json', size=bs)],
        [sg.Button('Get Plain Text from JSON', key='json_to_txt', size=bs)],
        [sg.Button('Markdown to TEI', key='md_to_tei', size=bs)],
        [sg.Button('TEI to JSON', key='tei_to_json', size=bs)],
        [sg.Button('Combine Collation Files', key='combine_verses', size=bs)],
        [sg.Button('Reformat Collation File', key='reformat_xml', size=bs)],
        [sg.Button('View TEI Transcriptions', key='tei_server', size=bs)],
        [sg.Button('Configure Collation Editor', key='ce_config', size=bs)],
        [sg.Button('open-cbgm Interface', key='open-cbgm', size=bs)],
        [sg.Button('Export Collation to DOCX', key='export_to_docx', size=bs)],
        [sg.Stretch(), sg.Button('Close', pad=(20, 20), size=(20, 1)), sg.Stretch()]
    ]
    return layout



from pathlib import Path
import platform

import PySimpleGUI as sg
#pylint: disable=import-error
# pylint: disable=no-member
import criticus.py.edit_settings as es 
import criticus.py.cbgm_interface.open_cbgm_api as oc
import criticus.py.custom_popups as cp

# operating_system = platform.system(


def validate_user_input(values: dict):
    if values['xml_file'] == '':
        cp.ok('Select a Collation file with which to populate the new database.', 'Forget something?')
        return
    elif not Path(values['xml_file']).is_file() or not values['xml_file'].endswith('.xml'):
        cp.ok('The given XML Collation file could not be opened or does not exist.', 'Cannot Open File')
        return
    if values['new_db_name'] == '':
        cp.ok('Enter a name for the new database', 'Forget something?')
        return
    if values['threshold'] and values['threshold_input'] == '':
        cp.ok('You have enabled readings threshold without providing\nthe threshold number.', 'Forget something?')
        return
    elif values['threshold']:
        try:
            int(values['threshold_input'])
        except:
            cp.ok('The reads threshold must be an integer', 'Bad Input')
            return
    if values['trivial'] and values['trivial_input'] == '':
        cp.ok('You have enabled "Treat as Trivial" without designating\nwhich reading types to trivialize.', 'Forget something?')
        return
    if values['exclude'] and values['exclude_input'] == '':
        cp.ok('You have chosen to "Exclude" reading type(s) without designating\nthe types to be excluded.', 'Forget something?')
        return
    return True

def validate_compare_inputs(values):
    if values['selected_db'] in ['', [], None]:
        return
    if values['wit_to_compare'] == '':
        cp.ok('Enter a witness to which others are compared.', 'Forget something?')
        return
    if values['compare_some'] and values['wits_to_compare'] == '':
        cp.ok('You have chosen to compare select witnesses without designating\nthose witnesses.', 'Forget something?')
        return
    return True

def populate_new_database(values, window: sg.Window):
    if not validate_user_input(values):
        return
    es.edit_settings('cbgm_main_dir', values['cbgm_main_dir'])
    es.edit_settings('cbgm_cx_dir', Path(values['xml_file']).parent.as_posix())
    oc.populate_db(values)
    window['db_listbox'].update(oc.get_all_dbs())
    window['selected_db'].update(oc.get_all_dbs())

def delete_selected(values, window):
    if values['db_listbox'] == []:
        return
    oc.delete_db(values)
    window['db_listbox'].update(oc.get_all_dbs())
    window['selected_db'].update(oc.get_all_dbs())

def compare_witnesses(values, window):
    if not validate_compare_inputs(values):
        return
    table = oc.compare_wits(values)
    if not table:
        cp.ok(f'There was a problem opening the database or locating witness "{values["wit_to_compare"]}".', 'Bummer')
        return None
    display_table = []
    for row in table['rows']:
        r = []
        for item in ['id', 'dir', 'pass', 'eq', 'perc', 'prior', 'posterior']:
            r.append(str(row[item]))
        display_table.append(r)
    window['table'].update(values=display_table)
    return display_table

def save_as_csv(values, settings):
    output_fn = sg.popup_get_file(
        '', save_as=True, file_types=(('CSV Files', '*.csv'),),
        no_window=True, initial_folder=settings['cbgm_main_dir'])
    if not output_fn:
        return
    result = oc.csv_comparison(values, output_fn)
    if result:
        cp.ok(f'The CSV file was saved to\n{output_fn}', 'Success!')
    else:
        cp.ok('CSV file was not saved.', 'Bummer')

def view_plain_text(values):
    text = oc.view_plain_text(values)
    cp.textbox(text, 'open-cbgm text output')

def manage_db_tab_layout(settings: dict):
    w = (25, 1)
    options_frame = [
        [sg.Checkbox('Readings Threshold: ', key='threshold', size=w), sg.I('', key='threshold_input')],
        [sg.Checkbox('Treat as Trivial: ', key='trivial', size=w), sg.I('', key='trivial_input')],
        [sg.Checkbox('Ignore: ', key='exclude', size=w), sg.I('', key='exclude_input')],
        [sg.Checkbox('Merge Split Attestations', key='merge_split')],
        [sg.T('Rules: '), sg.Radio('Standard', 'rules', default=True, key='standard'), sg.Radio('Classic', 'rules', key='classic')],
    ]
    xml_file_frame = [
        [sg.T('Select XML Collation File: ', size=w), sg.I('', key='xml_file'), sg.FileBrowse(file_types=(('XML Files', '*.xml'),), initial_folder=settings['cbgm_cx_dir'])],
        [sg.T('New Database Name: ', size=w), sg.I('', key='new_db_name')],
        
    ]
    dbs_frame = [
        [sg.Listbox(oc.get_all_dbs(), select_mode=sg.SELECT_MODE_EXTENDED, key='db_listbox', expand_x=True, expand_y=True)],
        [sg.B('Delete Selected')]
    ]
    return [
        [sg.T('open-cbgm binaries folder: '), sg.I(settings['cbgm_main_dir'], key='cbgm_main_dir'), sg.FolderBrowse(initial_folder=settings['cbgm_main_dir'])],
        [sg.Frame('Populate New Database', xml_file_frame)],
        [sg.Frame('Optional Settings', options_frame)],
        [sg.B('Populate Database')],
        [sg.Frame('All Databases', dbs_frame, expand_x=True, expand_y=True)],
    ]
def compare_wits_tab_layout(settings: dict):
    headings = ['Witness', 'Direction', 'Passages', 'Agreement', 'Percentage', 'Prior', 'Posterior']
    return [
        [sg.T('Select a Database'), sg.Combo(oc.get_all_dbs(), readonly=True, key='selected_db', expand_x=True)],
        [sg.T('Compare Witnesses to: '), sg.I('', key='wit_to_compare')],
        [sg.Radio('Compare All Witnesses', 'compare', default=True, key='compare_all')],
        [sg.Radio('Compare only these: ', 'compare', key='compare_some'), sg.I('', key='wits_to_compare')],
        [sg.B('Compare', bind_return_key=True)],
        [sg.Table([['','','','','','','',]], headings=headings, key='table', expand_x=True, expand_y=True)],
        [sg.B('Save as CSV'), sg.B('View Plain Text')],
    ]

def layout(settings):
    return [
        [sg.TabGroup([[sg.Tab('Manage Databases', manage_db_tab_layout(settings), expand_x=True, expand_y=True), sg.Tab('Compare Witnesses', compare_wits_tab_layout(settings), expand_x=True, expand_y=True)]], enable_events=True, key='tab', expand_x=True, expand_y=True)],
        [sg.B('Back', key='exit'), sg.Stretch()],
    ]

def open_cbgm_ui(font, icon):
    settings = es.get_settings()
    window = sg.Window('open-cbgm Interface', layout(settings), font=font, icon=icon, resizable=True)

    while True:
        event, values = window.read()
        if event in [sg.WINDOW_CLOSED, None, 'exit']:
            break

        elif event == 'Populate Database':
            populate_new_database(values, window)

        elif event == 'Delete Selected':
            delete_selected(values, window)

        elif event == 'Compare':
            compare_witnesses(values, window)

        elif event == 'Save as CSV':
            save_as_csv(values, settings)

        elif event == 'View Plain Text':
            view_plain_text(values)

        elif event == 'tab':
            window['selected_db'].update(values=oc.get_all_dbs())    
        
    window.close()
    return False



import codecs

import PySimpleGUI as sg

import criticus.py.edit_settings as es


def unescape_string(text: str):
    return codecs.decode(text, 'unicode-escape')

def update_window(settings: dict, window: sg.Window):
    window['regexes'].update(settings['pre_parse_regex'])
    window['regex'].update('')
    window['replacement'].update('')

def add_regex(values: dict, window: sg.Window):
    settings = es.get_settings()
    regex = unescape_string(values['regex'])
    replacement = unescape_string(values['replacement'])
    settings['pre_parse_regex'].append(
        [regex, replacement]
    )
    update_window(settings, window)
    es.save_settings(settings)

def delete_selected(values: dict, window: sg.Window):
    settings = es.get_settings()
    new_regexes = []
    for reg in settings['pre_parse_regex']:
        if reg in values['regexes']:
            continue
        new_regexes.append(reg)
    settings['pre_parse_regex'] = new_regexes
    update_window(settings, window)
    es.save_settings(settings)

def create_layout(regexes: list = []):
    return [
        [sg.T('Regular Expression: '), sg.I('', key='regex'), sg.T('for'), sg.I('', key='replacement'), sg.B('Add')],
        [sg.Listbox(regexes, select_mode=sg.SELECT_MODE_EXTENDED, key='regexes', expand_x=True, expand_y=True)],
        [sg.B('Delete Selected')],
        [sg.B('Done')]
    ]

def edit_regex(icon):
    settings = es.get_settings()
    layout = create_layout(settings['pre_parse_regex'])

    window = sg.Window('Add/Remove Regular Expressions', layout, icon=icon, resizable=True)

    while True:
        event, values = window.read()
        if event in [sg.WIN_CLOSED, None, 'Done']:
            break
        elif event == 'Add':
            add_regex(values, window)
        elif event == 'Delete Selected':
            delete_selected(values, window)

    window.close()



import random
import re
import xml.etree.ElementTree as et
from lxml.etree import _Element # pylint: disable=no-name-in-module
import lxml.etree as ET
from typing import List

from markdown.core import Markdown
from markdown.extensions import Extension
from markdown.preprocessors import Preprocessor
from markdown.treeprocessors import Treeprocessor


def preprocess_md(md: str):
    for x in punctuation:
        md = md.replace(x, f'<pc>{x}</pc>')
    md = md.replace('-\n', '') # this replacement must come before the <lb> sub
    md = md.replace('[', '~~')
    md = md.replace(']', '~~')
    md = re.sub(r'\n<lb([^<>]*)> *', r'\n<lb\1> ', md)
    return md

###########################################################
# post process markup before final xml processing.
# This section must result in valid xml to be parsed with lxml

def add_text_body_opening_tags(markup: str):
    '''User defines in markdown where the TEI header information ends
    and the transcription begins by placing three or more fullstops
    (e.g. '...') on their own line. Only replace first occurance so that
    the use of fullstops is available to user for other things (e.g. lacunae).'''
    return re.sub(r'\.\.\.+', '<text xml:lang="grc"><body>', markup, count=1)

def h4_to_book_div(markup: str):
    '''Repurpose the h4 tags renderd from the '#### <book>' markup syntax
    for a div tag that wraps all transcription pertaining to the book
    specified in the h4 text. User most close this tag by adding '####'
    without any text at the end of the book transcription.'''
    book_match = re.search(r'<h4>[^<>]+</h4>', markup).group(0)
    book = book_match.replace('<h4>', '')
    book = book.replace('</h4>', '')
    try:
        book = nt_to_igntp[book]
    except KeyError:
        pass
    book_elem = f'<div type="book" n="{book}">\n'
    markup = markup.replace(book_match, book_elem)
    return markup, book

def h5_to_chapter_div(markup: str, book):
    '''Repurpose the h5 tags renderd from the '##### <chapter number>' markup syntax
    for a div tag that wraps all transcription pertaining to the chapter
    specified in the h5 text. User most close this tag by adding '#####'
    without any text at the end of the chapter transcription.'''
    chapters = re.findall(r'<h5>[^<>]+</h5>', markup)
    for chapter_match in chapters:
        chapter = chapter_match.replace('<h5>', '')
        chapter = chapter.replace('</h5>', '')
        chapter_elem = f'<div type="chapter" n="{book}K{chapter}">'
        markup = markup.replace(chapter_match, chapter_elem)
    return markup

def bulk_replace(markup: str):
    '''Convert Markdown tag names to TEI tags
    (and other things).'''
    for x in html_to_tei:
        markup = markup.replace(x[0], x[1])
    return markup

def add_tei_boilerplate_and_ending_tags(markup: str):
    return f'''<!DOCTYPE TEI><?xml-stylesheet type="text/xsl" href="tei_transcription.xsl"?><TEI xmlns="http://www.tei-c.org/ns/1.0">{markup}</body></text></TEI>'''

def postprocess_markup(markup: str):
    markup = add_text_body_opening_tags(markup)
    markup, book = h4_to_book_div(markup)
    markup = h5_to_chapter_div(markup, book)
    markup = bulk_replace(markup)
    markup = bulk_replace(markup)
    markup = add_tei_boilerplate_and_ending_tags(markup)

    return markup

###########################################################
###########################################################
# Post process by parsing xml

def fill_out_untranscribed_commentary_markup(root: _Element):
    '''convert abbreviated "<comm/>" and "<comm lines=4> tags into IGNTP equivalent markup'''
    tei_ns = 'http://www.tei-c.org/ns/1.0'
    xml_ns = 'http://www.w3.org/XML/1998/namespace'
    comms = root.xpath(f'//tei:comm', namespaces={'tei': tei_ns}) # type: List[_Element]
    for comm in comms:
        if comm.get('lines'):
            index = comm.getparent().index(comm)
            try:
                lines = int(comm.get('lines'))
            except:
                lines = 1
            for _ in range(lines):
                parent = comm.getparent() #type: _Element
                comm_element = ET.Element('note', type='commentary', nsmap={None: tei_ns, 'xml': xml_ns})
                comm_element.text = 'One line of untranscribed commentary text'
                lb = ET.Element('lb', nsmap={None: tei_ns, 'xml': xml_ns})
                parent.insert(index, lb)
                parent.insert(index+1, comm_element)
            comm.getparent().remove(comm)
        else:
            comm.tag = 'note'
            comm.attrib['type'] = 'commentary'
            comm.text = 'untranscribed commentary text'
    return root

def fill_out_verse_unit_attributes(root: _Element):
    '''Convert verse number e.g. n="6" to full IGNTP format e.g. n="B06K11V6"'''
    tei_ns = 'http://www.tei-c.org/ns/1.0'
    verse_units = root.xpath(f'//tei:ab', namespaces={'tei': tei_ns}) # type: List[_Element]
    for v in verse_units: #type: _Element
        v.attrib['n'] = f'{v.getparent().get("n")}V{v.get("n")}'
    return root

def add_page_break_type(root: _Element):
    '''adds type="folio" to the page break element
    for a very minor convenvience'''
    tei_ns = 'http://www.tei-c.org/ns/1.0'
    for pb in root.xpath('//tei:pb', namespaces={'tei': tei_ns}):
        pb.attrib['type'] = 'folio'
    return root

def postprocess_xml(root: _Element):
    root = fill_out_untranscribed_commentary_markup(root)
    root = fill_out_verse_unit_attributes(root)
    root = add_page_break_type(root)
    return root

###########################################################
###########################################################

class TokenizeText(Preprocessor):
    """ wrap individual transcribed words in <w> tags """
    def rando_str(self):
        chars = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        random_list = [random.choice(chars) for _ in range(5)]
        rando =  ''.join(random_list)
        return f'$$${rando}'

    def replace_words_with_mid_break(self, text, replacements):
        breaks_in_word = re.findall(r'<lb[^<>]*>', text)
        for brk in breaks_in_word:
            rando = self.rando_str()
            text = text.replace(brk, f'{rando}')
            replacements.append((brk, rando))
        return text, replacements

    def replace_user_markups(self, text, replacements):
        user_markups = re.findall(r'{[^{}]+}', text)
        for m in user_markups:
            rando = self.rando_str()
            text = text.replace(m, f'{rando} ')
            replacements.append((m, rando))
        return text, replacements

    def replace_markups(self, text, replacements):
        markups = re.findall(r'<[^<>]+>', text)
        for m in markups:
            rando = self.rando_str()
            text = text.replace(m, f' {rando} ')
            replacements.append((m, f'{rando}'))
        return text, replacements

    def replace_notes(self, text, replacements):
        notes = re.findall(r'\*[^*]+\*', text)
        for n in notes:
            rando = self.rando_str()
            text = text.replace(n, f'{rando}')
            replacements.append((n, f'{rando}'))
        return text, replacements

    def build_new_text(self, text):
        new_text = []
        for word in text.split():
            if (word not in punctuation_no_spaces
                and word != '|'
                and word[0] not in ['*', '+']
                and not word.startswith('$$$')):

                word = word.replace(word, f'<w>{word}</w>')
            new_text.append(word)
        text = ''.join(new_text)
        return text

    def replace_randos(self, text, replacements):
        for r in replacements:
            text = text.replace(r[1], r[0])
        return text

    def tokenize_text(self, text: str):
        '''tokenize transcription words but ignore other 
        words (e.g. notes) and markup that may contain 
        spaces but should not be tokenized'''
        replacements = []
        text, replacements = self.replace_words_with_mid_break(text, replacements)
        text, replacements = self.replace_user_markups(text, replacements)
        text, replacements = self.replace_markups(text, replacements)
        text, replacements = self.replace_notes(text, replacements)   
        text = self.build_new_text(text)
        text = self.replace_randos(text, replacements)
        return text

    def run(self, lines: list):
        new_lines = []
        for line in lines:
            if line.startswith('#') or line.startswith('...') or line.startswith('>'):
                new_lines.append(line)
            else:
                new_line = self.tokenize_text(line)
                new_lines.append(new_line)
        return new_lines

class RestructureTree(Treeprocessor):
    def create_elements(self):
        return (
            et.Element('teiHeader'),
            et.Element('fileDesc'),
            et.Element('titleStmt'),
            et.Element('respStmt'),
            et.Element('name'),
            et.Element('resp')
        )
    def remove_elements(self, root, elements_to_remove: tuple):
        for elem in elements_to_remove:
            root.remove(elem)

    def get_header_info(self, root):
        return (
            root.find('h1'),
            root.find('h2'),
            root.find('h3'),
        )

    def build_tei_header(self, root):
        (tei_header, file_desc, title_statement, resp_statement, 
         name, resp) = self.create_elements()
        title, person, date = self.get_header_info(root)
        root.insert(0, tei_header)
        tei_header.append(file_desc)
        file_desc.append(title_statement)
        title_statement.append(title)
        title_statement.append(resp_statement)
        resp_statement.append(resp)
        resp_statement.append(name)
        title.attrib = {'type': 'document'}
        name.text = person.text
        name.attrib = {'type': 'person'}
        resp.text = 'Transcribed by'
        resp.attrib = {'when-iso': date.text}
        self.remove_elements(root, (title, date, person))

    def run(self, root: et.Element):
        self.build_tei_header(root)         


class TEI(Extension):
    def extendMarkdown(self, md: Markdown, key="TEI", index=200):
        md.treeprocessors.register(RestructureTree(md), 'RestructureTree', 200)
        md.preprocessors.register(TokenizeText(md), 'tokenize', 200)


punctuation = (' . ', ' , ', ' \u0387 ', ' : ', ' + ', ' ~ ')
punctuation_no_spaces = ('.', ',', '\u0387', ':', '+', '~')

html_to_tei = (
    ('<h1', '<title'),
    ('</h1>', '</title>'),
    ('<h2', '<respStmt'),
    ('</h2>', '</respStmt>'),
    ('<code', '<unclear'),
    ('</code>', '</unclear>'),
    ('<v', '<ab'),
    ('</v>', '</ab>'),
    ('<p>', ''),
    ('</p>', ''),
    ('<p/>', ''),
    ('<h4/>', '</div>'),
    ('<h4></h4>', '</div>'),
    ('<h5/>', '</div>'),
    ('<h5></h5>', '</div>'),
    ('<strong', '<seg type="margin" '),
    ('</strong>', '</seg>'),
    ('<del>', '<supplied>'),
    ('</del>', '</supplied>'),
    ('<em', '<note type="local"'),
    ('</em>', '</note>'),
    ('<ins>', '<app><rdg type="orig" hand="firsthand">'),
    ('|', '</rdg><rdg type="corr" hand="corrector">'),
    ('</ins>', '</rdg></app>'),
    ('\n', ''),
)

nt_to_igntp = {
    'Matt': 'B01',
    'Matthew': 'B01',
    'Mark': 'B02',
    'Luke': 'B03',
    'John': 'B04',
    'Acts': 'B05',
    'Rom': 'B06',
    'Romans': 'B06',
    '1 Cor': 'B07',
    '1Cor': 'B07',
    '1 Corinthians': 'B07',
    '2 Cor': 'B08',
    '2Cor': 'B08',
    '2 Corinthians': 'B08',
    'Gal': 'B09',
    'Galatians': 'B09',
    'Eph': 'B10',
    'Ephesians': 'B10',
    'Phil': 'B11',
    'Philippians': 'B11',
    'Col': 'B12',
    'Colossians': 'B12',
    '1 Thess': 'B13',
    '1Thess': 'B13',
    '1 Thessalonians': 'B13',
    '2 Thess': 'B14',
    '2Thess': 'B14',
    '2 Thessalonians': 'B14',
    '1 Tim': 'B15',
    '1Tim': 'B15',
    '1 Timothy': 'B15',
    '2 Tim': 'B16',
    '2Tim': 'B16',
    '2 Timothy': 'B16',
    'Titus': 'B17',
    'Phlm': 'B18',
    'Philemon': 'B18',
    'Heb': 'B19',
    'Hebrews': 'B19',
    'Jas': 'B20',
    'James': 'B20',
    '1 Pet': 'B21',
    '1Pet': 'B21',
    '1 Peter': 'B21',
    '2 Pet': 'B22',
    '2Pet': 'B22',
    '2 Peter': 'B22',
    '1 John': 'B23',
    '1John': 'B23',
    '2 John': 'B24',
    '2John': 'B24',
    '3 John': 'B25',
    '3John': 'B25',
    'Jude': 'B26',
    'Rev': 'B27',
    'Revelation': 'B27',
}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['criticus',
 'criticus.py',
 'criticus.py.cbgm_interface',
 'criticus.py.export_to_docx',
 'criticus.py.md2tei',
 'criticus.py.reformat_collation',
 'criticus.py.serve_tei_transcriptions',
 'criticus.py.tei2json',
 'criticus.py.txt2json',
 'criticus.resources']

package_data = \
{'': ['*']}

install_requires = \
['Markdown==3.2.2',
 'PySimpleGUI>=4.57.0,<5.0.0',
 'lxml>=4.8.0,<5.0.0',
 'markdown-del-ins>=1.0.0,<2.0.0',
 'natsort>=8.1.0,<9.0.0',
 'python-docx>=0.8.11,<0.9.0',
 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'criticus',
    'version': '0.35.2',
    'description': 'A suite of tools for transcribing, collating and creating an apparatus criticus.',
    'long_description': '# Criticus\nA suite of tools for transcribing, collating and and creating an apparatus criticus.\n\n## install\nRequires Python 3.8+\n\n### Windows: \n- `pip install criticus`\n- `python -m criticus`\n### MacOS: \n- `pip3 install criticus`\n- `python3 -m criticus`\n\nRead the full [tutorial on GitHub](https://github.com/d-flood/criticus).\n',
    'author': 'David Flood',
    'author_email': 'davidfloodii@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/d-flood/criticus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from lxml import etree as et

def version_rdgs(elem: et._Element, regularized: dict):
    if elem.get('n') in regularized:
        regularized[elem.get('n')] += 1
    else:
        regularized[elem.get('n')] = 1
    unique_name = f'{elem.get("n")}{regularized[elem.get("n")]}'
    elem.attrib['n'] = unique_name
    return regularized, unique_name

def differentiate_subreading_ids(xml_filename) -> et._ElementTree:
    ns = '{http://www.tei-c.org/ns/1.0}'
    parser = et.XMLParser(remove_blank_text=True)
    xml = et.parse(xml_filename, parser=parser)
    root = xml.getroot()
    for ab in root:
        for app in ab:
            regularized = {}
            for elem in app:
                if elem.tag == f'{ns}rdg' and 'r' in elem.get('n'):
                    regularized, _ = version_rdgs(elem, regularized)

                elif elem.tag == f'{ns}note': # this is a <note> element
                    for child in elem: # this is a <note> child
                        if child.tag == f'{ns}graph':
                            regularized = {}
                            for node in child:
                                if node.tag == f'{ns}node':
                                    if 'r' in node.get('n'):
                                        regularized, distinct_attrib = version_rdgs(node, regularized)
                                        arc = et.Element('arc')
                                        arc.attrib['from'] = distinct_attrib[0]
                                        arc.attrib['to'] = distinct_attrib
                                        node.getparent().append(arc)
    # xml.write(xml_filename, encoding='utf8')
    return xml

# def rename_orphan_subreadings(xml_filename):
#     ns = '{http://www.tei-c.org/ns/1.0}'
#     parser = et.XMLParser(remove_blank_text=True)
#     xml = et.parse(xml_filename, parser=parser)
#     root = xml.getroot()
#     for rdg in root.xpath('tei:rdg', namespaces={'tei': ns}):
#         if



import toml

from criticus.py import mureq as req
from criticus.py import custom_popups as cp

def check_for_update(current):
    response = req.get('https://raw.githubusercontent.com/d-flood/criticus/master/pyproject.toml')
    data = response.content.decode('utf-8')
    pyproject = toml.loads(data)
    newest = pyproject['tool']['poetry']['version']

    new_ver_message = f"""
There is a newer version of Criticus available!
Your current version is {current}.
The newest available is {newest}.
To upgrade, close Criticus, then copy the relavent 
line below and paste it into a terminal and hit 
Enter (Windows) or Return (MacOS).""".lstrip()

    if newest > current:
        cp.mac_win_cmd(new_ver_message, 'New Version Available', 'pip3 --upgrade criticus', 'pip --upgrade criticus')
    else:
        cp.ok('You already have the newest version of Criticus!', title="You're up to date")



import os
from pathlib import Path
import re
from typing import List

from lxml import etree as et
from docx import Document
from natsort import natsorted
import PySimpleGUI as sg

# from rich import print

from criticus.py.reformat_collation.itsee_to_open_cbgm import reformat_xml
import criticus.py.edit_settings as es

TEI_NS = '{http://www.tei-c.org/ns/1.0}'
XML_NS = '{http://www.w3.org/XML/1998/namespace}'
ABBR_TO_FULL = {
    'Matt': 'Matthew',
    'B01': 'Matthew',
    'Mark': 'Mark',
    'B02': 'Mark',
    'Luke': 'Luke',
    'B03': 'Luke',
    'John': 'John',
    'B04': 'John',
    'Acts': 'Acts',
    'B05': 'Acts',
    'Rom': 'Romans',
    'B06': 'Romans',
    'Romans': 'Romans',
    'R': 'Romans',
    '1 Cor': '1 Corinthians',
    '1Cor': '1 Corinthians',
    'ICor': '1 Corinthians',
    'B07': '1 Corinthians',
    'B07': '1 Corinthians',
    '1 Corinthians': '1 Corinthians',
    '2 Cor': '2 Corinthians',
    '2Cor': '2 Corinthians',
    'IICor': '2 Corinthians',
    '2 Corinthians': '2 Corinthians',
    'B08': '2 Corinthians',
    'Gal': 'Galatians',
    'Galatians': 'Galatians',
    'B09': 'Galatians',
    'Eph': 'Ephesians',
    'Ephesians': 'Ephesians',
    'B10': 'Ephesians',
    'Phil': 'Philippians',
    'Philippians': 'Philippians',
    'B11': 'Philippians',
    'Col': 'Colossians',
    'Colossians': 'Colossians',
    'B12': 'Colossians',
    '1 Thess': '1 Thessalonians',
    '1Thess': '1 Thessalonians',
    '1 Thessalonians': '1 Thessalonians',
    'B13': '1 Thessalonians',
    '2 Thess': '2 Thessalonians',
    '2Thess': '2 Thessalonians',
    '2 Thessalonians': '2 Thessalonians',
    'B14': '2 Thessalonians',
    '1 Tim': '1 Timothy',
    '1Tim': '1 Timothy',
    '1 Timothy': '1 Timothy',
    'B15': '1 Timothy',
    '2 Tim': '2 Timothy',
    '2Tim': '2 Timothy',
    '2 Timothy': '2 Timothy',
    'B16': '2 Timothy',
    'Titus': 'Titus',
    'B17': 'Titus',
    'Phlm': 'B18',
    'Philemon': 'Philemon',
    'B18': 'Philemon',
    'Heb': 'B19',
    'Hebrews': 'Hebrews',
    'B19': 'Hebrews',
    'Jas': 'James',
    'James': 'James',
    'B20': 'James',
    '1 Pet': '1 Peter',
    '1Pet': '1 Peter',
    '1 Peter': '1 Peter',
    'B21': '1 Peter',
    '2 Pet': '2 Peter',
    '2Pet': '2 Peter',
    '2 Peter': '2 Peter',
    'B22': '2 Peter',
    '1 John': '1 John',
    'B23': '1 John',
    '1John': '1 John',
    '2 John': '2 John',
    'B24': '2 John',
    '2John': '2 John',
    '3 John': '3John',
    '3John': '3John',
    'B25': '3John',
    'Jude': 'Jude',
    'B26': 'Jude',
    'Rev': 'Revelation',
    'Revelation': 'Revelation',
    'B27': 'Revelation',
}

def get_xml_file(xml: str) -> et._Element:
    temp_cx_file = 'temp_xml_collation_file'
    xml = xml.replace('xml:id="1', 'xml:id="I')
    xml = xml.replace('xml:id="2', 'xml:id="II')
    xml = xml.replace('xml:id="3', 'xml:id="III')
    xml = xml.replace('subreading', 'subr')
    with open(temp_cx_file, 'w', encoding='utf-8') as file:
        file.write(xml)
    if re.search('<teiHeader>', xml) is None:
        try:
            temp_cx_file = reformat_xml(temp_cx_file)
        except:
            return None
    parser = et.XMLParser(remove_blank_text=True, encoding='UTF-8')
    tree = et.parse(temp_cx_file, parser) #type: et._ElementTree
    root = tree.getroot()
    os.remove(temp_cx_file)
    return root

def get_document():
    this_dir = Path(__file__).parent
    template = this_dir.joinpath('template.docx').as_posix()
    # print(template)
    return Document(template)

def get_custom_document(filepath: str):
    try:
        return Document(filepath)
    except:
        if sg.popup_yes_no('Criticus failed to open the custom template.\nWould you like to use the default template instead?', title='Failed to Load Custom Template') == 'Yes':
            return get_document()
        else:
            return

def load_xml_file(xml_file: str):
    with open(xml_file, 'r', encoding='utf-8') as file:
        xml = file.read()
    return get_xml_file(xml)

def construct_full_ref(ab: et. _Element):
    ref = ab.get(f'{XML_NS}id').replace('-APP', '').upper() #type: str
    if ref.startswith('B'): # then it is an INTF/IGNTP style reference... probably
        book = re.search(r'B\d+', ref).group(0)
        book = ABBR_TO_FULL[book]
        chapter = re.search(r'K\d+', ref).group(0)
        verse = re.search(r'V\d+', ref).group(0)
        ref = f'{book} {chapter.replace("K", "")}:{verse.replace("V", "")}'
    else:
        book = re.search(r'.[a-zA-Z]+', ref)
        if not book:
            return ref
        book = book.group(0)
        full_book = ABBR_TO_FULL.get(book)
        if not full_book:
            return ref
        reference = ref.replace(book, '').replace('.', ':')
        ref = f'{full_book} {reference}'
    return ref

def print_reference(document: Document, ab: et._Element):
    ref = construct_full_ref(ab)
    reference = document.add_paragraph(ref)
    reference.style = document.styles['reference']

def group_basetext_words(basetext: str, words_per_line: int) -> List[list]:
    words_per_line = words_per_line - 1
    grouped_basetext = []
    current_group = []
    chunk = 0
    for word in basetext.split():
        if chunk == words_per_line:
            current_group.append(word)
            grouped_basetext.append(current_group)
            chunk = 0
            current_group = []
            continue
        current_group.append(word)
        chunk += 1
    if current_group != []:
        grouped_basetext.append(current_group)
    return grouped_basetext

def construct_basetext(ab: et._Element) -> str:
    basetext = []
    for elem in ab:
        if elem.tag == f'{TEI_NS}seg':
            basetext.append(elem.text)
        elif elem.tag == f'{TEI_NS}app' and elem.find(f'{TEI_NS}lem').get('type') != 'om':
            basetext.append(elem.find(f'{TEI_NS}lem').text)
    return ' '.join(basetext)

def print_basetext(document: Document, ab: et._Element, words_per_line: int):
    basetext = construct_basetext(ab)
    basetext = group_basetext_words(basetext, words_per_line)
    table = document.add_table(rows=0, cols=10)
    index = 2
    for line in basetext:
        row_cells = table.add_row().cells
        for cell, word in enumerate(line):
            row_cells[cell].text = f"{word}\n{index}"
            row_cells[cell].paragraphs[0].style = document.styles['table cell']
            index += 2

def print_app(document: Document, app: et._Element):
    app_from = app.get('from')
    app_to = app.get('to')
    if app_from == app_to:
        index = app_from
    else:
        index = f'{app_from}–{app_to}'
    p = document.add_paragraph(index)
    p.style = document.styles['index']

def sort_by_ga(wits: List[str]):
    papyri = []
    majuscules = []
    minuscules = []
    lectionaries = []
    editions = []
    for wit in wits:
        # try:
        if wit.lower().startswith('p'):
            papyri.append(wit)
        elif wit.startswith('0'):
            majuscules.append(wit)
        elif wit[0].isdigit():
            minuscules.append(wit)
        elif wit.lower().startswith('l'):
            lectionaries.append(wit)
        else:
            editions.append(wit)
        # except:
        #     print(wit)
    return natsorted(papyri) + natsorted(majuscules) + natsorted(minuscules) + natsorted(lectionaries) + natsorted(editions)

# TODO: format wits same as Apparatus Explorer
# def format_wits(wits: str):
#     wits = re.sub(r'\([^()]\)', '', wits)
#     wits = wits.split()
#     wits = sort_by_ga(wits)
    

def print_rdg(
    document, rdg: et._Element, 
    text_wits_separator: str, 
    rdg_n_text_separator: str, 
    text_bold: bool
    ):
    if rdg.get('type') and rdg.text:
        greek_text = f"{rdg.get('type')}\t{rdg.text}"
    elif rdg.get('type'):
        greek_text = rdg.get('type')
    else:
        greek_text = rdg.text
    # if rdg.text:
    #     greek_text = rdg.text
    # else:
    #     greek_text = rdg.get('type')
    p = document.add_paragraph()
    p.style = document.styles['reading']
    rdg_name = re.sub(r'\d', '', rdg.get('n'))
    p.add_run(rdg_name).italic = True
    p.add_run(rdg_n_text_separator)
    p.add_run(greek_text).bold = text_bold
    wits = rdg.get('wit')

    wits = sort_by_ga(wits)
    wits = ' '.join(wits)
    p.add_run(f"{text_wits_separator}{wits}")

def save_docx(document: Document, settings: dict):
    docx_filename = sg.popup_get_file('', no_window=True, save_as=True, initial_folder=settings.get('export_docx_folder'), file_types=(('DOCX Files', '*.docx'),))
    if not docx_filename:
        return
    if not docx_filename.endswith('.docx'):
        docx_filename = f'{docx_filename}.docx'
    docx_dir = Path(docx_filename).parent.as_posix()
    es.edit_settings('export_docx_folder', docx_dir)
    try:
        document.save(docx_filename)
    except PermissionError:
        sg.popup_ok('You cannot overwrite a file that is currently open in another app.', 'Oopsie')
    return docx_filename

def combine_regularized(app: et._Element, add_suffix: bool):
    reg_wits = {}
    for rdg in app.findall(f'{TEI_NS}rdg'): #type: List[et._Element]
        rdg_id = rdg.get('n')
        if 'r' not in rdg_id:
            parent = rdg_id
        elif rdg_id[0] != parent[0]:
            continue
        else:
            reg_wits[parent] = f"{reg_wits.get(parent, '')} {rdg.get('wit')}".lstrip()

    for parent, wits in reg_wits.items():
        wits = wits.split()
        if add_suffix:
            wits = [f'{w  }r' for w in wits]
        wits = ' '.join(wits)
        reg_wits[parent] = wits

    previous = ''
    for rdg in app.findall(f'{TEI_NS}rdg'): #type: list[et._Element]
        if rdg.get('n') in reg_wits:
            combined = f'{rdg.get("wit")} {reg_wits[rdg.get("n")]}'.lstrip()
            rdg.attrib['wit'] = combined
        elif 'r' in rdg.get('n') and rdg.get('n')[0] != previous[0]:
            continue
        elif 'r' in rdg.get('n'):
            app.remove(rdg)
        previous = rdg.get('n')

    return app

def export_xml_to_docx(xml_filename: str, collapse_regularized: bool = False, use_custom_template: str = None, add_suffix: bool = False):
    settings = es.get_settings()
    if use_custom_template:
        document = get_custom_document(use_custom_template)
    else:
        document = get_document()
    if not document:
        return
    root = load_xml_file(xml_filename)
    for ab in root.findall(f'{TEI_NS}ab'):
        print_reference(document, ab)
        print_basetext(document, ab, settings['words_per_line'])
        for app in ab.findall(f'{TEI_NS}app'):
            if collapse_regularized:
                app = combine_regularized(app, add_suffix)
            if len(app.findall(f'{TEI_NS}rdg')) == 1:
                continue

            print_app(document, app)
            for rdg in app.findall(f'{TEI_NS}rdg'): #type: List[et._Element]
                print_rdg(
                    document, rdg, settings['text_wits_separator'], 
                    settings['rdg_n_text_separator'], settings['text_bold']
                    )

    return save_docx(document, settings)
